# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['DiscreteSymmetry']

import numpy as np
from scipy.sparse import identity, csr_matrix, hstack


def almost_identity(mat):
    return np.all(abs(mat - identity(mat.shape[0])).data < 1e-10)


def _column_sum(matrix):
    """Sum the columns of a sparse matrix.

    This is fully analogous to ``matrix.sum(0)``, and uses an implementation
    similar to that in scipy v1.1.0, however it avoids using ``numpy.matrix``
    interface and therefore does not raise a ``PendingDeprecationWarning``.
    This should be removed once we depend on scipy v1.2.0, where the warning is
    silenced.
    """
    return matrix.T @ np.ones(matrix.shape[0])


def cond_conj(op, conj):
    return op.conj() if conj else op


_conj = [True, True, False]
_names = ['Time reversal', 'Particle-hole', 'Chiral']
_signs = [1, -1, -1]

class DiscreteSymmetry:
    r"""A collection of discrete symmetries and conservation laws.

    Parameters
    ----------
    projectors : iterable of rectangular or square sparse matrices
        Projectors that block-diagonalize the Hamiltonian.
    time_reversal : square sparse matrix
        The unitary part of the time-reversal symmetry operator.
    particle_hole : square sparse matrix
        The unitary part of the particle-hole symmetry operator.
    chiral : square sparse matrix
        The chiral symmetry operator.

    Notes
    -----

    When computing scattering modes, the representation of the
    modes is chosen to reflect declared discrete symmetries and
    conservation laws.

    `projectors` block diagonalize the Hamiltonian, and modes are computed
    separately in each block. The ordering of blocks is the same as of
    `projectors`. If `conservation_law` is declared in
    `~kwant.builder.Builder`, `projectors` is computed as the projectors
    onto its orthogonal eigensubspaces. The projectors are stored in the
    order of ascending eigenvalues of `conservation_law`.

    Symmetrization using discrete symmetries varies depending on whether
    a conservation law/projectors are declared. Consider the case with no
    conservation law declared. With `time_reversal` declared, the outgoing
    modes are chosen as the time-reversed partners of the incoming modes,
    i.e. :math:`\psi_{out}(-k) = T \psi_{in}(k)` with k the momentum.
    `chiral` also relates incoming and outgoing modes, such that
    :math:`\psi_{out}(k) = C \psi_{in}(k)`. `particle_hole` gives symmetric
    incoming and outgoing modes separately, such that
    :math:`\psi_{in/out}(-k) = P \psi_{in/out}(k)`, except when k=-k, at
    k = 0 or :math:`\pi`. In this case, each mode is chosen as an eigenstate
    :math:`P \psi = \psi` if :math:`P^2=1`. If :math:`P^2=-1`, we
    symmetrize the modes by generating pairs of orthogonal modes
    :math:`\psi` and :math:`P\psi`. Because `chiral` and `particle_hole`
    flip the sign of energy, they only apply at zero energy.

    Discrete symmetries can be combined with a conservation law if they
    leave each block invariant or transform it to another block. With S
    a discrete symmetry and :math:`P_i` and :math:`P_j` projectors onto
    blocks i and j of the Hamiltonian, :math:`S_{ji} = P_j^+ S P_i` is the
    symmetry projection that maps from block i to block j.
    :math:`S_{ji} = P_j^+ S P_i` must for each j be nonzero for exactly
    one i. If S leaves block i invariant, the modes within block i are
    symmetrized using the nonzero projection :math:`S_{ii}`, like in the
    case without a conservation law. If S transforms between blocks i and
    j, the modes of the block with the larger index are obtained by
    transforming the modes of the block with the lower index. Thus, with
    :math:`\psi_i` and :math:`\psi_j` the modes of blocks i and j, we have
    :math:`\psi_j = S_{ji} \psi_i`.
    """
    def __init__(self, projectors=None, time_reversal=None, particle_hole=None,
                 chiral=None):
        # Normalize format
        if projectors is not None:
            try:
                projectors = [proj.tocsr() for proj in projectors]
            except AttributeError:
                raise TypeError("projectors must be a sequence of "
                                "sparse matrices.")

        symms = (time_reversal, particle_hole, chiral)
        try:
            symms = [symm.tocsr() if symm is not None else None
                     for symm in symms]
        except AttributeError:
            raise TypeError("Symmetries must be sparse matrices.")

        if None not in (time_reversal, particle_hole, chiral):
            product = chiral.dot(particle_hole).dot(time_reversal.conj())
            if not almost_identity(product):
                raise ValueError("Product of all symmetries "
                                 "should be equal to identity.")

        # Auto-compute a missing symmetry.
        if None not in (time_reversal, particle_hole) and chiral is None:
            chiral = particle_hole.dot(time_reversal.conj())
        elif None not in (particle_hole, chiral) and time_reversal is None:
            # Product of particle hole with itself to get the sign right.
            time_reversal = (particle_hole.dot(particle_hole.conj()).dot(
                             particle_hole).dot(chiral.conj()))
        elif None not in (chiral, time_reversal) and particle_hole is None:
            particle_hole = (time_reversal.dot(time_reversal.conj()).dot(
                             chiral).dot(time_reversal))

        symms = [time_reversal, particle_hole, chiral]
        for i, symm, conj, name in zip(range(3), symms, _conj, _names):
            if symm is None:
                continue
            if not almost_identity(symm.T.conj().dot(symm)):
                raise ValueError("{} symmetry should be "
                                 "{}unitary.".format(name, "anti" * conj))
            symm_squared = symm.dot(cond_conj(symm, conj))
            if not (almost_identity(symm_squared) or
                    (conj and almost_identity(-symm_squared))):
                raise ValueError("{} symmetry should square to "
                                 "{}1.".format(name, "±" * conj))
            symms[i] = symm

        if projectors is not None:
            projectors = [
                p[:, _column_sum(abs(p)) > 1e-10] for p in projectors
            ]
            if not almost_identity(sum(projector.dot(projector.conj().T) for
                                       projector in projectors)):
                raise ValueError("Projectors should stack to a unitary matrix.")

            # Check that the symmetries and conservation law are in canonical
            # form.  In the projector basis, each discrete symmetry should have
            # at most one nonzero block per row.
            for (symm, name, conj) in zip(symms, _names, _conj):
                if symm is None:
                    continue
                for p_i in projectors:
                    p_hc = p_i.T.conj()
                    nonzero_blocks = sum(
                            np.any(abs(p_hc.dot(symm).dot(
                                (cond_conj(p_j, conj))).data > 1e-7))
                                         for p_j in projectors)
                    if nonzero_blocks > 1:
                        raise ValueError(name + ' symmetry is not in canonical '
                                         'form: see DiscreteSymmetry '
                                         'documentation.')
        self.projectors = projectors
        self.time_reversal, self.particle_hole, self.chiral = symms

    def validate(self, matrix):
        """Check if a matrix satisfies the discrete symmetries.

        Parameters
        ----------
        matrix : sparse or dense matrix
            If rectangular, it is padded by zeros to be square.

        Returns
        -------
        broken_symmetries : list
            List of strings, the names of symmetries broken by the
            matrix: any combination of "Conservation law", "Time reversal",
            "Particle-hole", "Chiral". If no symmetries are broken, returns
            an empty list.
        """
        # Extra transposes are to enforse sparse dot product in case matrix is
        # dense.
        n, m = matrix.shape
        if n != m:
            if isinstance(matrix, np.ndarray):
                new_matrix = np.empty((n, n), dtype=matrix.dtype)
                new_matrix[:, :m] = matrix
                new_matrix[:, m:] = 0
                matrix = new_matrix
            else:
                matrix = hstack([matrix, csr_matrix((n, n-m))])
        broken_symmetries = []
        if self.projectors is not None:
            for proj in self.projectors:
                full = proj.dot(proj.T.conj())
                commutator = full.dot(matrix) - (full.T.dot(matrix.T)).T
                if np.linalg.norm(commutator.data) > 1e-8:
                    broken_symmetries.append('Conservation law')
                    break
        for symm, conj, sign, name in zip(self[1:], _conj, _signs, _names):
            if symm is None:
                continue
            commutator = symm.T.conj().dot((symm.T.dot(matrix.T)).T)
            commutator = commutator - sign * cond_conj(matrix, conj)
            if np.linalg.norm(commutator.data) > 1e-8:
                broken_symmetries.append(name)
        return broken_symmetries

    def __getitem__(self, item):
        return (self.projectors, self.time_reversal,
                self.particle_hole, self.chiral)[item]



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import numpy as np
from pytest import raises
from numpy.testing import assert_almost_equal
import kwant
from kwant.physics import two_terminal_shotnoise
from kwant._common import ensure_rng

n = 5
chain = kwant.lattice.chain()

def twoterminal_system():
    rng = ensure_rng(11)
    system = kwant.Builder()
    lead = kwant.Builder(kwant.TranslationalSymmetry((1,)))
    h = rng.random_sample((n, n)) + 1j * rng.random_sample((n, n))
    h += h.conjugate().transpose()
    h *= 0.8
    t = 4 * rng.random_sample((n, n)) + 4j * rng.random_sample((n, n))
    lead[chain(0)] = h
    system[chain(0)] = h * 1.2
    lead[chain(0), chain(1)] = t
    system.attach_lead(lead)
    system.attach_lead(lead.reversed())
    return system


def test_multiterminal_input():
    """Input checks for two_terminal_shotnoise"""

    syst = twoterminal_system()
    syst.attach_lead(syst.leads[0].builder)
    sol = kwant.smatrix(syst.finalized(), out_leads=[0], in_leads=[0])
    raises(ValueError, two_terminal_shotnoise, sol)


def test_twoterminal():
    """Shot noise in a two-terminal conductor"""

    fsyst = twoterminal_system().finalized()

    sol = kwant.smatrix(fsyst)
    t = sol.submatrix(1, 0)
    Tn = np.linalg.eigvalsh(np.dot(t, t.conj().T))
    noise_should_be = np.sum(Tn * (1 - Tn))

    assert_almost_equal(noise_should_be, two_terminal_shotnoise(sol))



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""Band structure calculation for the leads"""

import math
import numpy as np
from .. import system
from .._common import ensure_isinstance, deprecate_args

__all__ = ['Bands']


class Bands:
    """
    Class of callable objects for the computation of energy bands.

    Parameters
    ----------
    sys : `kwant.system.InfiniteSystem`
        The low level infinite system for which the energies are to be
        calculated.
    args : tuple, defaults to empty
        Positional arguments to pass to the ``hamiltonian`` method.
        Deprecated in favor or 'params' (and mutually exclusive with it).
    params : dict, optional
        Dictionary of parameter names and their values. Mutually exclusive
        with 'args'.

    Notes
    -----
    An instance of this class can be called like a function.  Given a momentum
    (currently this must be a scalar as all infinite systems are quasi-1-d), it
    returns a NumPy array containing the eigenenergies of all modes at this
    momentum. If `derivative_order > 0` or `return_eigenvectors = True`,
    additional arrays are returned.

    Examples
    --------
    >>> bands = kwant.physics.Bands(some_syst)
    >>> momenta = numpy.linspace(-numpy.pi, numpy.pi, 101)
    >>> energies = [bands(k) for k in momenta]
    >>> pyplot.plot(momenta, energies)
    >>> pyplot.show()
    """
    _crossover_size = 8

    @deprecate_args
    def __init__(self, sys, args=(), *, params=None):
        syst = sys
        ensure_isinstance(syst, system.InfiniteSystem)
        self.ham = syst.cell_hamiltonian(args, params=params)
        if not np.allclose(self.ham, self.ham.T.conj()):
            raise ValueError('The cell Hamiltonian is not Hermitian.')
        hop = syst.inter_cell_hopping(args, params=params)
        self.hop = np.empty(self.ham.shape, dtype=complex)
        self.hop[:, : hop.shape[1]] = hop
        self.hop[:, hop.shape[1]:] = 0

    def __call__(self, k, derivative_order=0, return_eigenvectors=False):
        r"""Calculate band energies at a given momentum.

        :math:`E_n, \quad E'_n = dE_n / dk, \quad E''_n = d^2E_n / dk^2,
        \quad n \in \{0, nbands - 1\}`

        :math:`nbands` is the number of open modes.
        Eigenvectors are orthonormal.

        Parameters
        ----------
        k : float
            momentum
        derivative_order : {0, 1, 2}, optional
            Maximal derivative order to calculate. Default is zero.
        return_eigenvectors : bool, optional
            if set to `True` return the eigenvectors as last tuple element.
            By default, no eigenvectors are returned.

        Returns
        -------
        energies : numpy float array, size ``nbands``
            energies :math:`E`
        velocities : numpy float array, size ``nbands``
            velocities (first order energy derivatives :math:`E'`)
        curvatures : numpy float array, size ``nbands``
            curvatures (second energy derivatives :math:`E''`)
        eigenvectors : numpy float array, size ``nbands x nbands``
            eigenvectors

        The number of returned elements varies. If ``derivative_order > 0``
        we return all derivatives up to the order ``derivative_order``,
        and total the number of returned elements is ``derivative_order + 1``
        If ``return_eigenvectors`` is True, in addition the eigenvectors are
        returned as the last element. In that case, the total number of
        returned elements is ``derivative_order + 2``.

        Notes
        -----
        * All the output arrays are sorted from the lowest energy band
          to the highest.
        * The curvature `E''` can be only calculated for non-degenerate bands.
        """
        # Equation to solve is
        # (V^\dagger e^{ik} + H + V e^{-ik}) \psi = E \psi
        # we obtain the derivatives by perturbing around momentum k
        # The perturbed Hamiltonian H(k) is
        # H(k+dk) = H_0(k) + H_1(k) dk + H_2(k) dk^2 + O(dk^3)
        # and we use h0, h1 and h2 in the code to refer to above terms

        mat = self.hop * complex(math.cos(k), -math.sin(k))
        h0 = mat + mat.conjugate().transpose() + self.ham

        # compute energies and if required eigenvectors
        # numpy routines eigh and eigvalsh return eigenvalues in ascending order
        if return_eigenvectors or derivative_order > 0:
            energies, eigenvectors = np.linalg.eigh(h0)
        else:
            energies = np.linalg.eigvalsh(h0)
        output = (energies.real,)

        if derivative_order >= 1:  # compute velocities
            h1 = 1j*(- mat + mat.conjugate().transpose())
            # Switch computation from diag(a^\dagger @ b) to sum(a.T * b, axis=0),
            # which is computationally more efficient for larger matrices.
            # a and b are NxN matrices, and we switch for N > _crossover_size.
            # Both ways are numerically (but not necessarily bitwise) equivalent.
            if derivative_order == 1 and len(energies) > self._crossover_size:
                velocities = np.sum(eigenvectors.conjugate() * (h1 @ eigenvectors),
                                    axis=0).real
            else:
                ph1p = eigenvectors.conjugate().transpose() @ h1 @ eigenvectors
                velocities = np.diag(ph1p).real
            output += (velocities,)

        if derivative_order >= 2:  # compute curvatures
            # ediff_{i,j} =  1 / (E_i - E_j) if i != j, else 0
            ediff = energies.reshape((-1, 1)) - energies.reshape((1, -1))
            ediff = np.divide(1, ediff, where=(ediff != 0))
            h2 = - (mat + mat.conjugate().transpose())
            curvatures = np.sum(eigenvectors.conjugate() * (h2 @ eigenvectors)
                                - 2 * ediff * np.abs(ph1p)**2, axis=0).real
            # above expression is similar to: curvatures =
            # np.diag(eigenvectors.conjugate().transpose() @ h2 @ eigenvectors
            #         + 2 * ediff @ np.abs(ph1p)**2).real
            output += (curvatures,)

        if derivative_order > 2:
            raise NotImplementedError('Derivatives of the energy dispersion ' +
                                      'only implemented up to second order.')
        if return_eigenvectors:
            output += (eigenvectors,)
        if len(output) == 1:
            # Backwards compatibility: if returning only energies,
            # don't make it a length-1 tuple.
            return output[0]
        return output



#  tests for kwant-1.4.2-py37hc8548de_0 (this is a generated file);
print('===== testing package: kwant-1.4.2-py37hc8548de_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import sys
import os.path

from pytest import main

import kwant

def test():
    kwant_path = os.path.dirname(os.path.abspath(kwant.__file__))
    sys.exit(main([kwant_path, "-x", "-v"]))

test()
#  --- run_test.py (end) ---

print('===== kwant-1.4.2-py37hc8548de_0 OK =====');
print("import: 'kwant'")
import kwant

print("import: 'kwant.graph'")
import kwant.graph

print("import: 'kwant.graph.tests'")
import kwant.graph.tests

print("import: 'kwant.linalg'")
import kwant.linalg

print("import: 'kwant.linalg.tests'")
import kwant.linalg.tests

print("import: 'kwant.physics'")
import kwant.physics

print("import: 'kwant.physics.tests'")
import kwant.physics.tests

print("import: 'kwant.solvers'")
import kwant.solvers

print("import: 'kwant.solvers.tests'")
import kwant.solvers.tests

print("import: 'kwant.tests'")
import kwant.tests




# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from kwant.graph import Graph
# from kwant.graph.scotch import bisect, reset

def _DISABLED_test_bisect():
    # REMARK: This test is somewhat limited in the sense that it can only test
    #         for the general sanity of the output, not if the bisection is
    #         really good (balanced, etc.).
    size = 5
    graph = Graph()

    for i in range(size - 1):
        offset = i * size
        for j in range(size - 1):
            graph.add_edge(offset + j, offset + j + 1)
            graph.add_edge(offset + j + 1, offset + j)
        if i > 0:
            for j in range(size):
                graph.add_edge(offset + j, offset + j - size)
                graph.add_edge(offset + j - size, offset + j)
    g = graph.compressed()

    parts = bisect(g)
    for i in range(g.num_nodes):
        assert parts[i] == 0 or parts[i] == 1

def _DISABLED_test_reset():
    size = 5
    graph = Graph()

    for i in range(size - 1):
        offset = i * size
        for j in range(size - 1):
            graph.add_edge(offset + j, offset + j + 1)
            graph.add_edge(offset + j + 1, offset + j)
        if i > 0:
            for j in range(size):
                graph.add_edge(offset + j, offset + j - size)
                graph.add_edge(offset + j - size, offset + j)
    g = graph.compressed()

    # After calling reset, SCOTCH returns identical results.
    reset()
    parts1 = bisect(g)
    reset()
    parts2 = bisect(g)

    assert (parts1 == parts2).all()



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


from math import sin, cos, sqrt, pi, copysign
from collections import namedtuple

from itertools import combinations_with_replacement
import numpy as np
import numpy.linalg as npl
import scipy.linalg as la
from .. import linalg as kla
from scipy.linalg import block_diag
from scipy.sparse import (identity as sp_identity, hstack as sp_hstack,
                          csr_matrix)

dot = np.dot

__all__ = ['selfenergy', 'modes', 'PropagatingModes', 'StabilizedModes']


# TODO: Use scipy block_diag once we depend on scipy>=0.19
try:
    # Throws ValueError, but if fixed ensure that works as intended
    bdiag_broken = block_diag(np.zeros((1,1)), np.zeros((2,0))).shape != (3, 1)
except ValueError:  # skip coverage
    bdiag_broken = True
if bdiag_broken:  # skip coverage
    def block_diag(*matrices):
        """Construct a block diagonal matrix out of the input matrices.

        Like scipy.linalg.block_diag, but works for zero size matrices."""
        rows, cols = np.sum([mat.shape for mat in matrices], axis=0)
        b_mat = np.zeros((rows,cols), dtype='complex')
        rows, cols = 0, 0
        for mat in matrices:
            new_rows = rows + mat.shape[0]
            new_cols = cols + mat.shape[1]
            b_mat[rows:new_rows, cols:new_cols] = mat
            rows, cols = new_rows, new_cols
        return b_mat


# TODO: Remove the workaround once we depend on scipy >= 1.0
def lstsq(a, b):
    """Least squares version that works also with 0-shaped matrices."""
    if a.shape[1] == 0:
        return np.empty((0, 0), dtype=np.common_type(a, b))
    return la.lstsq(a, b)[0]


def nonzero_symm_projection(matrix):
    """Check whether a discrete symmetry relation between two blocks of the
    Hamiltonian vanishes or not.

    For a discrete symmetry S, the projection that maps from the block i to
    block j of the Hamiltonian with projectors P_i and P_j is
    S_ji = P_j^+ S P_i.
    This function determines whether S_ji vanishes or not, i. e. whether a
    symmetry relation exists between blocks i and j.

    Here, discrete symmetries and projectors are in canonical form, such that
    S maps each block at most either to itself or to a single other block.
    In other words, for each j, the block S_ji is nonzero at most for one i,
    while all other blocks vanish. If a nonzero block exists, it is unitary
    and hence has norm 1. To identify nonzero blocks without worrying about
    numerical errors, we thus check that its norm is larger than 0.5.
    """
    if not isinstance(matrix, np.ndarray):
        matrix = matrix.data
    return np.linalg.norm(matrix) > 0.5


def group_halves(arr_list):
    """Split and rearrange a list of arrays.

    Combine a list of arrays into a single array where the first halves
    of each array appear first:
    `[a b], [c d], [e f] -> [a c e b d f]`
    """
    list_ = [np.split(arr, 2) for arr in arr_list]
    lefts, rights = zip(*list_)
    return np.r_[tuple(lefts + rights)]


# Container classes
Linsys = namedtuple('Linsys', ['eigenproblem', 'v', 'extract'])


class PropagatingModes:
    """The calculated propagating modes of a lead.

    Attributes
    ----------
    wave_functions : numpy array
        The wave functions of the propagating modes.
    momenta : numpy array
        Momenta of the modes.
    velocities : numpy array
        Velocities of the modes.
    block_nmodes: list of integers
        Number of left or right moving propagating modes
        per conservation law block of the Hamiltonian.

    Notes
    =====
    The sort order of all the three arrays is identical. The first half of the
    modes have negative velocity, the second half have positive velocity.
    Within these halves the modes are ordered by the eigenvalues of any
    declared conservation laws. Within blocks with the same conservation law
    eigenvalue the modes with negative velocity are ordered by increasing
    momentum, and the modes with positive velocity are ordered by decreasing
    momentum. Finally, modes are ordered by the magnitude of their velocity.
    To summarize, the modes are ordered according to the key
    `(sign(v), conserved_quantity, sign(v) * k , abs(v))` where `v` is
    velocity, `k` is momentum and `conserved_quantity` is the conservation
    law eigenvalue.

    In the above, the positive velocity and momentum directions are defined
    with respect to the translational symmetry direction of the system.

    The first dimension of `wave_functions` corresponds to the orbitals of all
    the sites in a unit cell, the second one to the number of the mode.  Each
    mode is normalized to carry unit current. If several modes have the same
    momentum and velocity, an arbitrary orthonormal basis in the subspace of
    these modes is chosen.

    If a conservation law is specified to block diagonalize the Hamiltonian,
    then `block_nmodes[i]` is the number of left or right moving propagating
    modes in conservation law block `i`. The ordering of blocks is the same as
    the ordering of the projectors used to block diagonalize the Hamiltonian.
    """
    def __init__(self, wave_functions, velocities, momenta):
        kwargs = locals()
        kwargs.pop('self')
        self.__dict__.update(kwargs)


class StabilizedModes:
    """Stabilized eigendecomposition of the translation operator.

    Due to the lack of Hermiticity of the translation operator, its
    eigendecomposition is frequently poorly conditioned. Solvers in Kwant use
    this stabilized decomposition of the propagating and evanescent modes in
    the leads. If the hopping between the unit cells of an infinite system is
    invertible, the translation eigenproblem is written in the basis `psi_n,
    h_hop^+ * psi_(n+1)`, with ``h_hop`` the hopping between unit cells.  If
    `h_hop` is not invertible, and has the singular value decomposition `u s
    v^+`, then the eigenproblem is written in the basis `sqrt(s) v^+ psi_n,
    sqrt(s) u^+ psi_(n+1)`. In this basis we calculate the eigenvectors of the
    propagating modes, and the Schur vectors (an orthogonal basis) in the space
    of evanescent modes.

    `vecs` and `vecslmbdainv` are the first and the second halves of the wave
    functions.  The first `nmodes` are eigenmodes moving in the negative
    direction (hence they are incoming into the system in Kwant convention),
    the second `nmodes` are eigenmodes moving in the positive direction. The
    remaining modes are the Schur vectors of the modes evanescent in the
    positive direction. Propagating modes with the same eigenvalue are
    orthogonalized, and all the propagating modes are normalized to carry unit
    current. Finally the `sqrt_hop` attribute is `v sqrt(s)`.

    Attributes
    ----------
    vecs : numpy array
        Translation eigenvectors.
    vecslmbdainv : numpy array
        Translation eigenvectors divided by the corresponding eigenvalues.
    nmodes : int
        Number of left-moving (or right-moving) modes.
    sqrt_hop : numpy array or None
        Part of the SVD of `h_hop`, or None if the latter is invertible.
    """

    def __init__(self, vecs, vecslmbdainv, nmodes, sqrt_hop=None):
        kwargs = locals()
        kwargs.pop('self')
        self.__dict__.update(kwargs)

    def selfenergy(self):
        """
        Compute the self-energy generated by lead modes.

        Returns
        -------
        Sigma : numpy array, real or complex, shape (M,M)
            The computed self-energy. Note that even if `h_cell` and `h_hop` are
            both real, `Sigma` will typically be complex. (More precisely, if
            there is a propagating mode, `Sigma` will definitely be complex.)
        """
        v = self.sqrt_hop
        vecs = self.vecs[:, self.nmodes:]
        vecslmbdainv = self.vecslmbdainv[:, self.nmodes:]
        return dot(v, dot(vecs, la.solve(vecslmbdainv, v.T.conj())))


# Auxiliary functions that perform different parts of the calculation.
def setup_linsys(h_cell, h_hop, tol=1e6, stabilization=None):
    """Make an eigenvalue problem for eigenvectors of translation operator.

    Parameters
    ----------
    h_cell : numpy array with shape (n, n)
        Hamiltonian of a single lead unit cell.
    h_hop : numpy array with shape (n, m), m <= n
        Hopping Hamiltonian from a cell to the next one.
    tol : float
        Numbers are considered zero when they are smaller than `tol` times
        the machine precision.
    stabilization : sequence of 2 booleans or None
        Which steps of the eigenvalue problem stabilization to perform. If the
        value is `None`, then Kwant chooses the fastest (and least stable)
        algorithm that is expected to be sufficient.  For any other value,
        Kwant forms the eigenvalue problem in the basis of the hopping singular
        values.  The first element set to `True` forces Kwant to add an
        anti-Hermitian term to the cell Hamiltonian before inverting. If it is
        set to `False`, the extra term will only be added if the cell
        Hamiltonian isn't invertible. The second element set to `True` forces
        Kwant to solve a generalized eigenvalue problem, and not to reduce it
        to the regular one.  If it is `False`, reduction to a regular problem
        is performed if possible.

    Returns
    -------
    linsys : namedtuple
        A named tuple containing `matrices` a matrix pencil defining
        the eigenproblem, `v` a hermitian conjugate of the last matrix in
        the hopping singular value decomposition, and functions for
        extracting the wave function in the unit cell from the wave function
        in the basis of the nonzero singular exponents of the hopping.

    Notes
    -----
    The lead problem with degenerate hopping is rather complicated, and the
    details of the algorithm will be published elsewhere.

    """
    n = h_cell.shape[0]
    m = h_hop.shape[1]
    if stabilization is not None:
        stabilization = list(stabilization)

    if not np.any(h_hop):  # skip coverage
        # Inter-cell hopping is zero.  The current algorithm is not suited to
        # treat this extremely singular case.
        raise ValueError("Inter-cell hopping is exactly zero.")

    # If both h and t are real, it may be possible to use the real eigenproblem.
    if (not np.any(h_hop.imag)) and (not np.any(h_cell.imag)):
        h_hop = h_hop.real
        h_cell = h_cell.real

    eps = np.finfo(np.common_type(h_cell, h_hop)).eps * tol

    # First check if the hopping matrix has singular values close to 0.
    # (Close to zero is defined here as |x| < eps * tol * s[0] , where
    # s[0] is the largest singular value.)

    u, s, vh = la.svd(h_hop)
    assert m == vh.shape[1], "Corrupt output of svd."
    n_nonsing = np.sum(s > eps * s[0])

    if (n_nonsing == n and stabilization is None):
        # The hopping matrix is well-conditioned and can be safely inverted.
        # Hence the regular transfer matrix may be used.
        h_hop_sqrt = sqrt(np.linalg.norm(h_hop))
        A = h_hop / h_hop_sqrt
        B = h_hop_sqrt
        B_H_inv = 1.0 / B     # just a real scalar here
        A_inv = la.inv(A)

        lhs = np.zeros((2*n, 2*n), dtype=np.common_type(h_cell, h_hop))
        lhs[:n, :n] = -dot(A_inv, h_cell) * B_H_inv
        lhs[:n, n:] = -A_inv * B
        lhs[n:, :n] = A.T.conj() * B_H_inv

        def extract_wf(psi, lmbdainv):
            return B_H_inv * np.copy(psi[:n])

        matrices = (lhs, None)
        v_out = h_hop_sqrt * np.eye(n)
    else:
        if stabilization is None:
            stabilization = [None, False]

        # The hopping matrix has eigenvalues close to 0 - those
        # need to be eliminated.

        # Recast the svd of h_hop = u s v^dagger such that
        # u, v are matrices with shape n x n_nonsing.
        u = u[:, :n_nonsing]
        s = s[:n_nonsing]
        u = u * np.sqrt(s)
        # pad v with zeros if necessary
        v = np.zeros((n, n_nonsing), dtype=vh.dtype)
        v[:vh.shape[1]] = vh[:n_nonsing].T.conj()
        v = v * np.sqrt(s)

        # Eliminating the zero eigenvalues requires inverting the on-site
        # Hamiltonian, possibly including a self-energy-like term.  The
        # self-energy-like term stabilizes the inversion, but the most stable
        # choice is inherently complex. This can be disadvantageous if the
        # Hamiltonian is real, since staying in real arithmetics can be
        # significantly faster.  The strategy here is to add a complex
        # self-energy-like term always if the original Hamiltonian is complex,
        # and check for invertibility first if it is real

        matrices_real = issubclass(np.common_type(h_cell, h_hop), np.floating)
        add_imaginary = stabilization[0] or ((stabilization[0] is None) and
                                             not matrices_real)
        # Check if there is a chance we will not need to add an imaginary term.
        if not add_imaginary:
            h = h_cell
            sol = kla.lu_factor(h)
            rcond = kla.rcond_from_lu(sol, npl.norm(h, 1))

            if rcond < eps:
                need_to_stabilize = True
            else:
                need_to_stabilize = False

        if add_imaginary or need_to_stabilize:
            need_to_stabilize = True
            # Matrices are complex or need self-energy-like term to be
            # stabilized.
            temp = dot(u, u.T.conj()) + dot(v, v.T.conj())
            h = h_cell + 1j * temp

            sol = kla.lu_factor(h)
            rcond = kla.rcond_from_lu(sol, npl.norm(h, 1))

            # If the condition number of the stabilized h is
            # still bad, there is nothing we can do.
            if rcond < eps:
                raise RuntimeError("Flat band encountered at the requested "
                                   "energy, result is badly defined.")

        # The function that can extract the full wave function psi from
        # the projected one (v^dagger psi lambda^-1, s u^dagger psi).

        def extract_wf(psi, lmbdainv):
            wf = -dot(u, psi[: n_nonsing] * lmbdainv) - dot(v, psi[n_nonsing:])
            if need_to_stabilize:
                wf += 1j * (dot(v, psi[: n_nonsing]) +
                            dot(u, psi[n_nonsing:] * lmbdainv))
            return kla.lu_solve(sol, wf)

        # Setup the generalized eigenvalue problem.

        A = np.zeros((2 * n_nonsing, 2 * n_nonsing), np.common_type(h, h_hop))
        B = np.zeros((2 * n_nonsing, 2 * n_nonsing), np.common_type(h, h_hop))

        begin, end = slice(n_nonsing), slice(n_nonsing, None)

        A[end, begin] = np.identity(n_nonsing)
        temp = kla.lu_solve(sol, v)
        temp2 = dot(u.T.conj(), temp)
        if need_to_stabilize:
            A[begin, begin] = -1j * temp2
        A[begin, end] = temp2
        temp2 = dot(v.T.conj(), temp)
        if need_to_stabilize:
            A[end, begin] -= 1j *temp2
        A[end, end] = temp2

        B[begin, end] = -np.identity(n_nonsing)
        temp = kla.lu_solve(sol, u)
        temp2 = dot(u.T.conj(), temp)
        B[begin, begin] = -temp2
        if need_to_stabilize:
            B[begin, end] += 1j * temp2
        temp2 = dot(v.T.conj(), temp)
        B[end, begin] = -temp2
        if need_to_stabilize:
            B[end, end] = 1j * temp2

        v_out = v[:m]

        # Solving a generalized eigenproblem is about twice as expensive
        # as solving a regular eigenvalue problem.
        # Computing the LU factorization is negligible compared to both
        # (approximately 1/30th of a regular eigenvalue problem).
        # Because of this, it makes sense to try to reduce
        # the generalized eigenvalue problem to a regular one, provided
        # the matrix B can be safely inverted.

        lu_b = kla.lu_factor(B)
        if not stabilization[1]:
            rcond = kla.rcond_from_lu(lu_b, npl.norm(B, 1))
            # A more stringent condition is used here since errors can
            # accumulate from here to the eigenvalue calculation later.
            stabilization[1] = rcond > eps * tol

        if stabilization[1]:
            matrices = (kla.lu_solve(lu_b, A), None)
        else:
            matrices = (A, B)
    return Linsys(matrices, v_out, extract_wf)


def unified_eigenproblem(a, b=None, tol=1e6):
    """A helper routine for modes(), that wraps eigenproblems.

    This routine wraps the regular and general eigenproblems that can arise
    in a unfied way.

    Parameters
    ----------
    a : numpy array
        The matrix on the left hand side of a regular or generalized eigenvalue
        problem.
    b : numpy array or None
        The matrix on the right hand side of the generalized eigenvalue problem.
    tol : float
        The tolerance for separating eigenvalues with absolute value 1 from the
        rest.

    Returns
    -------
    ev : numpy array
        An array of eigenvalues (can contain NaNs and Infs, but those
        are not accessed in `modes()`) The number of eigenvalues equals
        twice the number of nonzero singular values of
        `h_hop` (so `2*h_cell.shape[0]` if `h_hop` is invertible).
    evanselect : numpy integer array
        Index array of right-decaying modes.
    propselect : numpy integer array
        Index array of propagating modes (both left and right).
    vec_gen(select) : function
        A function that computes the eigenvectors chosen by the array select.
    ord_schur(select) : function
        A function that computes the unitary matrix (corresponding to the right
        eigenvector space) of the (general) Schur decomposition reordered such
        that the eigenvalues chosen by the array select are in the top left
        block.
    """
    if b is None:
        eps = np.finfo(a.dtype).eps * tol
        t, z, ev = kla.schur(a)

        # Right-decaying modes.
        select = abs(ev) > 1 + eps
        # Propagating modes.
        propselect = abs(abs(ev) - 1) < eps

        vec_gen = lambda x: kla.evecs_from_schur(t, z, select=x)
        ord_schur = lambda x: kla.order_schur(x, t, z, calc_ev=False)[1]

    else:
        eps = np.finfo(np.common_type(a, b)).eps * tol
        s, t, z, alpha, beta = kla.gen_schur(a, b, calc_q=False)

        # Right-decaying modes.
        select = abs(alpha) > (1 + eps) * abs(beta)
        # Propagating modes.
        propselect = (abs(abs(alpha) - abs(beta)) < eps * abs(beta))

        with np.errstate(divide='ignore', invalid='ignore'):
            ev = alpha / beta
        # Note: the division is OK here, since we later only access
        #       eigenvalues close to the unit circle

        vec_gen = lambda x: kla.evecs_from_gen_schur(s, t, z=z, select=x)
        ord_schur = lambda x: kla.order_gen_schur(x, s, t, z=z,
                                                  calc_ev=False)[2]

    return ev, select, propselect, vec_gen, ord_schur


def phs_symmetrization(wfs, particle_hole):
    """Makes the wave functions that have the same velocity at a time-reversal
    invariant momentum (TRIM) particle-hole symmetric.

    If P is the particle-hole operator and P^2 = 1, then a particle-hole
    symmetric wave function at a TRIM is an eigenstate of P with eigenvalue 1.
    If P^2 = -1, wave functions with the same velocity at a TRIM come in pairs.
    Such a pair is particle-hole symmetric if the wave functions are related by
    P, i. e. the pair can be expressed as [psi_n, P psi_n] where psi_n is a wave
    function.

    To ensure proper ordering of modes, this function also returns an array
    of indices which ensures that particle-hole partners are properly ordered
    in this subspace of modes. These are later used with np.lexsort to ensure
    proper ordering.

    Parameters
    ----------
    wfs : numpy array
        A matrix of propagating wave functions at a TRIM that all have the same
        velocity. The orthonormal wave functions form the columns of this matrix.
    particle_hole : numpy array
        The matrix representation of the unitary part of the particle-hole
        operator, expressed in the tight binding basis.

    Returns
    -------
    new_wfs : numpy array
        The matrix of particle-hole symmetric wave functions.
    TRIM_sort: numpy integer array
        Index array that stores the proper sort order of particle-hole symmetric
        wave functions in this subspace.
    """

    def Pdot(mat):
        """Apply the particle-hole operator to an array. """
        return particle_hole.dot(mat.conj())

    # Take P in the subspace of W = wfs: U = W^+ @ P @ W^*.
    U = wfs.T.conj().dot(Pdot(wfs))
    # Check that wfs are orthonormal and the space spanned
    # by them is closed under ph, meaning U is unitary.
    if not np.allclose(U.dot(U.T.conj()), np.eye(U.shape[0])):
        raise ValueError('wfs are not orthonormal or not closed under particle_hole.')
    P_squared = U.dot(U.conj())
    if np.allclose(P_squared, np.eye(U.shape[0])):
        P_squared = 1
    elif np.allclose(P_squared, -np.eye(U.shape[0])):
        P_squared = -1
    else:
        raise ValueError('particle_hole must square to +1 or -1. P_squared = {}'.format(P_squared))

    if P_squared == 1:
        # Use the matrix square root method from
        # Applied Mathematics and Computation 234 (2014) 380-384.
        assert np.allclose(U, U.T)
        # Schur decomposition guarantees that vecs are orthonormal.
        vals, vecs = la.schur(U)
        # U should be normal, so vals is diagonal.
        assert np.allclose(np.diag(np.diag(vals)), vals)
        vals = np.diag(vals)
        # Need to take safe square root of U, the branch cut should not go
        # through any eigenvalues. Otherwise the square root may not be symmetric.
        # Find largest gap between eigenvalues
        phases = np.sort(np.angle(vals))
        dph = np.append(np.diff(phases), phases[0] + 2*np.pi - phases[-1])
        i = np.argmax(dph)
        shift = -np.pi - (phases[i] + dph[i]/2)
        # Take matrix square root with branch cut in largest gap
        vals = np.sqrt(vals * np.exp(1j * shift)) * np.exp(-0.5j * shift)
        sqrtU = vecs.dot(np.diag(vals)).dot(vecs.T.conj())
        # For symmetric U sqrt(U) is also symmetric.
        assert np.allclose(sqrtU, sqrtU.T)
        # We want a new basis W_new such that W_new^+ @ P @ W_new^* = 1.
        # This is achieved by W_new = W @ sqrt(U).
        new_wfs = wfs.dot(sqrtU)
        # If P^2 = 1, there is no need to sort the modes further.
        TRIM_sort = np.zeros((wfs.shape[1],), dtype=int)
    else:
        # P^2 = -1.
        # Iterate over wave functions to construct
        # particle-hole partners.
        new_wfs = []
        # The number of modes. This is always an even number >=2.
        N_modes = wfs.shape[1]
        # If there are only two modes in this subspace, they are orthogonal
        # so we replace the second one with the P applied to the first one.
        if N_modes == 2:
            wf = wfs[:, 0]
            # Store psi_n and P psi_n.
            new_wfs.append(wf)
            new_wfs.append(Pdot(wf))
        # If there are more than two modes, iterate over wave functions
        # and construct their particle-hole partners one by one.
        else:
            # We construct pairs of modes that are particle-hole partners.
            # Need to iterate over all pairs except the final one.
            iterations = range((N_modes-2)//2)
            for i in iterations:
                # Take a mode psi_n from the basis - the first column
                # of the matrix of remaining modes.
                wf = wfs[:, 0]
                # Store psi_n and P psi_n.
                new_wfs.append(wf)
                P_wf = Pdot(wf)
                new_wfs.append(P_wf)
                # Remove psi_n and P psi_n from the basis matrix of modes.
                # First remove psi_n.
                wfs = wfs[:, 1:]
                # Now we project the remaining modes onto the orthogonal
                # complement of P psi_n. projector:
                projector = wfs.dot(wfs.T.conj()) - \
                            np.outer(P_wf, P_wf.T.conj())
                # After the projection, the mode matrix is rank deficient -
                # the span of the column space has dimension one less than
                # the number of columns.
                wfs = projector.dot(wfs)
                wfs = la.qr(wfs, mode='economic', pivoting=True)[0]
                # Remove the redundant column.
                wfs = wfs[:, :-1]
                # If this is the final iteration, we only have two modes
                # left and can construct particle-hole partners without
                # the projection.
                if i == iterations[-1]:
                    assert wfs.shape[1] == 2
                    wf = wfs[:, 0]
                    # Store psi_n and P psi_n.
                    new_wfs.append(wf)
                    new_wfs.append(Pdot(wf))
                assert np.allclose(wfs.T.conj().dot(wfs),
                                   np.eye(wfs.shape[1]))
        new_wfs = np.hstack([col.reshape(len(col), 1)/npl.norm(col) for
                             col in new_wfs])
        assert np.allclose(new_wfs[:, 1::2], Pdot(new_wfs[:, ::2]))
        # Store sort ordering in this subspace of modes
        TRIM_sort = np.arange(new_wfs.shape[1])
    assert np.allclose(new_wfs.T.conj().dot(new_wfs), np.eye(new_wfs.shape[1]))
    return new_wfs, TRIM_sort


def make_proper_modes(lmbdainv, psi, extract, tol, particle_hole,
                      time_reversal, chiral):
    """
    Find, normalize and sort the propagating eigenmodes.

    Special care is taken of the case of degenerate k-values, where the
    numerically computed modes are typically a superposition of the real
    modes. In this case, also the proper (orthogonal) modes are computed.
    """
    vel_eps = np.finfo(psi.dtype).eps * tol

    nmodes = psi.shape[1]
    n = len(psi) // 2

    # Array for the velocities.
    velocities = np.empty(nmodes, dtype=float)

    # Array of indices to sort modes at a TRIM by PHS.
    TRIM_PHS_sort = np.zeros(nmodes, dtype=int)

    # Calculate the full wave function in real space.
    full_psi = extract(psi, lmbdainv)

    # Cast the types if any of the symmetry operators is complex
    for symm in time_reversal, particle_hole, chiral:
        if symm is None:
            continue
        full_psi = full_psi.astype(np.common_type(symm, full_psi))
        psi = psi.astype(np.common_type(symm, psi))

    # Find clusters of nearby eigenvalues. Since the eigenvalues occupy the
    # unit circle, special care has to be taken to not introduce a cut at
    # lambda = -1.
    eps = np.finfo(lmbdainv.dtype).eps * tol
    angles = np.angle(lmbdainv)
    sort_order = np.resize(np.argsort(angles), (2 * len(angles,)))
    boundaries = np.argwhere(np.abs(np.diff(lmbdainv[sort_order]))
                             > eps).flatten() + 1

    # Detect the singular case of all eigenvalues equal.
    if boundaries.shape == (0,) and len(angles):
        boundaries = np.array([0, len(angles)])

    for interval in zip(boundaries[:-1], boundaries[1:]):
        if interval[1] > boundaries[0] + len(angles):
            break

        indx = sort_order[interval[0] : interval[1]]

        # If there is a degenerate eigenvalue with several different
        # eigenvectors, the numerical routines return some arbitrary
        # overlap of the real, physical solutions. In order
        # to figure out the correct wave function, we need to
        # have the full, not the projected wave functions
        # (at least to our current knowledge).

        # Finding the true modes is done in two steps:

        # 1. The true transversal modes should be orthogonal to each other, as
        # they share the same Bloch momentum (note that transversal modes with
        # different Bloch momenta k1 and k2 need not be orthogonal, the full
        # modes are orthogonal because of the longitudinal dependence e^{i k1
        # x} and e^{i k2 x}).  The modes with the same k are therefore
        # orthogonalized. Moreover for the velocity to have a proper value the
        # modes should also be normalized.

        q, r = la.qr(full_psi[:, indx], mode='economic')

        # If the eigenvectors were purely real up to this stage,
        # they will typically become complex after the rotation.
        if psi.dtype != np.common_type(psi, r):
            psi = psi.astype(np.common_type(psi, r))
        if full_psi.dtype != np.common_type(full_psi, q):
            full_psi = full_psi.astype(np.common_type(psi, q))

        full_psi[:, indx] = q
        psi[:, indx] = la.solve(r.T, psi[:, indx].T).T

        # 2. Moving infinitesimally away from the degeneracy
        # point, the modes should diagonalize the velocity
        # operator (i.e. when they are non-degenerate any more)
        # The modes are therefore rotated properly such that they
        # diagonalize the velocity operator.
        # Note that step 2. does not give a unique result if there are
        # two modes with the same velocity, or if the modes stay
        # degenerate even for a range of Bloch momenta (and hence
        # must have the same velocity). However, this does not matter,
        # since we are happy with any superposition in this case.

        vel_op = -1j * dot(psi[n:, indx].T.conj(), psi[:n, indx])
        vel_op = vel_op + vel_op.T.conj()
        vel_vals, rot = la.eigh(vel_op)

        # If the eigenvectors were purely real up to this stage,
        # they will typically become complex after the rotation.

        if psi.dtype != np.common_type(psi, rot):
            psi = psi.astype(np.common_type(psi, rot))
        if full_psi.dtype != np.common_type(full_psi, rot):
            full_psi = full_psi.astype(np.common_type(psi, rot))

        psi[:, indx] = dot(psi[:, indx], rot)
        full_psi[:, indx] = dot(full_psi[:, indx], rot)
        velocities[indx] = vel_vals

        # With particle-hole symmetry, treat TRIMs individually.
        # Particle-hole conserves velocity.
        # If P^2 = 1, we can pick modes at a TRIM as particle-hole eigenstates.
        # If P^2 = -1, a mode at a TRIM and its particle-hole partner are
        # orthogonal, and we pick modes such that they are related by
        # particle-hole symmetry.

        # At a TRIM, propagating translation eigenvalues are +1 or -1.
        if (particle_hole is not None and
            (np.abs(np.abs(lmbdainv[indx].real) - 1) < eps).all()):
            assert not len(indx) % 2
            # Set the eigenvalues to the exact TRIM values.
            if (np.abs(lmbdainv[indx].real - 1) < eps).all():
                lmbdainv[indx] = 1
            else:
                # Momenta are the negative arguments of the translation
                # eigenvalues, as computed below using np.angle. np.angle of -1
                # is pi, so this assigns k = -pi to modes with translation
                # eigenvalue -1.
                lmbdainv[indx] = -1

            # Original wave functions
            orig_wf = full_psi[:, indx]

            # Modes are already sorted by velocity in ascending order, as
            # returned by la.eigh above. The first half is thus incident,
            # and the second half outgoing.
            # Here we work within a subspace of modes with a fixed velocity.
            # Mostly, this is done to ensure that modes of different velocities
            # are not mixed when particle-hole partners are constructed for
            # P^2 = -1. First, we identify which modes have the same velocity.
            # In each such subspace of modes, we construct wave functions that
            # are particle-hole partners.
            vels = velocities[indx]
            # Velocities are sorted in ascending order. Find the indices of the
            # last instance of each unique velocity.
            inds = [ind+1 for ind, vel in enumerate(vels[:-1])
                    if np.abs(vel-vels[ind+1])>vel_eps]
            inds = [0] + inds + [len(vels)]
            inds = zip(inds[:-1], inds[1:])
            # Now possess an iterator over tuples, where each tuple (i,j)
            # contains the starting and final indices i and j of a submatrix
            # of the modes matrix, such that all modes in the submatrix
            # have the same velocity.

            # Iterate over all submatrices of modes with the same velocity.
            new_wf = []
            TRIM_sorts = []
            for ind_tuple in inds:
                # Pick out wave functions that have a given velocity
                wfs = orig_wf[:, slice(*ind_tuple)]
                # Make particle-hole symmetric modes
                new_modes, TRIM_sort = phs_symmetrization(wfs, particle_hole)
                new_wf.append(new_modes)
                # Store sorting indices of the TRIM modes with the given
                # velocity.
                TRIM_sorts.append(TRIM_sort)
            # Gather into a matrix of modes
            new_wf = np.hstack(new_wf)
            # Store the sort order of all modes at the TRIM.
            # Used later with np.lexsort when the ordering
            # of modes is done.
            TRIM_PHS_sort[indx] = np.hstack(TRIM_sorts)
            # Replace the old modes.
            full_psi[:, indx] = new_wf
            # For both cases P^2 = +-1, must rotate wave functions in the
            # singular value basis. Find the rotation from new basis to old.
            rot = new_wf.T.conj().dot(orig_wf)
            # Rotate the wave functions in the singular value basis
            psi[:, indx] = psi[:, indx].dot(rot.T.conj())

        # Ensure proper usage of chiral symmetry.
        if chiral is not None and time_reversal is None:
            out_orig = full_psi[:, indx[len(indx)//2:]]
            out = chiral.dot(full_psi[:, indx[:len(indx)//2]])
            # No least squares below because the modes should be orthogonal.
            rot = out_orig.T.conj().dot(out)
            full_psi[:, indx[len(indx)//2:]] = out
            psi[:, indx[len(indx)//2:]] = psi[:, indx[len(indx)//2:]].dot(rot)

    if np.any(abs(velocities) < vel_eps):
        raise RuntimeError("Found a mode with zero or close to zero velocity.")
    if 2 * np.sum(velocities < 0) != len(velocities):
        raise RuntimeError("Numbers of left- and right-propagating "
                           "modes differ, possibly due to a numerical "
                           "instability.")

    momenta = -np.angle(lmbdainv)
    # Sort the modes. The modes are sorted first by velocity and momentum,
    # and finally TRIM modes are properly ordered.
    order = np.lexsort([TRIM_PHS_sort, velocities,
                        -np.sign(velocities) * momenta, np.sign(velocities)])

    velocities = velocities[order]
    momenta = momenta[order]
    full_psi = full_psi[:, order]
    psi = psi[:, order]

    # Use particle-hole symmetry to relate modes that propagate in the
    # same direction but at opposite momenta.
    # Modes are sorted by velocity (first incident, then outgoing).
    # Incident modes are then sorted by momentum in ascending order,
    # and outgoing modes in descending order.
    # Adopted convention is to get modes with negative k (both in and out)
    # by applying particle-hole operator to modes with positive k.
    if particle_hole is not None:
        N = nmodes//2  # Number of incident or outgoing modes.
        # With particle-hole symmetry, N must be an even number.
        # Incident modes
        positive_k = (np.pi - eps > momenta[:N]) * (momenta[:N] > eps)
        # Original wave functions with negative values of k
        orig_neg_k = full_psi[:, :N][:, positive_k[::-1]]
        # For incident modes, ordering of wfs by momentum as returned by kwant
        # is [-k2, -k1, k1, k2], if k2, k1 > 0 and k2 > k1.
        # To maintain this ordering with ki and -ki as particle-hole partners,
        # reverse the order of the product at the end.
        wf_neg_k = particle_hole.dot(
                (full_psi[:, :N][:, positive_k]).conj())[:, ::-1]
        rot = lstsq(orig_neg_k, wf_neg_k)
        full_psi[:, :N][:, positive_k[::-1]] = wf_neg_k
        psi[:, :N][:, positive_k[::-1]] = \
                psi[:, :N][:, positive_k[::-1]].dot(rot)

        # Outgoing modes
        positive_k = (np.pi - eps > momenta[N:]) * (momenta[N:] > eps)
        # Original wave functions with negative values of k
        orig_neg_k = full_psi[:, N:][:, positive_k[::-1]]
        # For outgoing modes, ordering of wfs by momentum as returned by kwant
        # is like [k2, k1, -k1, -k2], if k2, k1 > 0 and k2 > k1.

        # Reverse order at the end to match momenta of opposite sign.
        wf_neg_k = particle_hole.dot(
                full_psi[:, N:][:, positive_k].conj())[:, ::-1]
        rot = lstsq(orig_neg_k, wf_neg_k)
        full_psi[:, N:][:, positive_k[::-1]] = wf_neg_k
        psi[:, N:][:, positive_k[::-1]] = \
                psi[:, N:][:, positive_k[::-1]].dot(rot)

    # Modes are ordered by velocity.
    # Use time-reversal symmetry to relate modes of opposite velocity.
    if time_reversal is not None:
        # Note: within this function, nmodes refers to the total number
        # of propagating modes, not either left or right movers.
        out_orig = full_psi[:, nmodes//2:]
        out = time_reversal.dot(full_psi[:, :nmodes//2].conj())
        rot = lstsq(out_orig, out)
        full_psi[:, nmodes//2:] = out
        psi[:, nmodes//2:] = psi[:, nmodes//2:].dot(rot)

    norm = np.sqrt(abs(velocities))
    full_psi = full_psi / norm
    psi = psi / norm

    return psi, PropagatingModes(full_psi, velocities, momenta)


def compute_block_modes(h_cell, h_hop, tol, stabilization,
                        time_reversal, particle_hole, chiral):
    """Calculate modes corresponding to a single projector. """
    n, m = h_hop.shape

    # Defer most of the calculation to helper routines.
    matrices, v, extract = setup_linsys(h_cell, h_hop, tol, stabilization)
    ev, evanselect, propselect, vec_gen, ord_schur = unified_eigenproblem(
        *(matrices + (tol,)))

    # v is never None.
    # h_hop.shape[0] and v.shape[1] not always the same.
    # Adding this makes the difference for some tests
    n = v.shape[1]

    nrightmovers = np.sum(propselect) // 2
    nevan = n - nrightmovers
    evan_vecs = ord_schur(evanselect)[:, :nevan]

    # Compute the propagating eigenvectors.
    prop_vecs = vec_gen(propselect)
    # Compute their velocity, and, if necessary, rotate them.

    # prop_vecs here is 'psi' in make_proper_modes, i.e. the wf in the SVD
    # basis.  It is in turn used to construct vecs and vecslmbdainv (the
    # propagating parts).  The evanescent parts of vecs and vecslmbdainv come
    # from evan_vecs above.
    prop_vecs, real_space_data = make_proper_modes(ev[propselect], prop_vecs,
                                                   extract, tol, particle_hole,
                                                   time_reversal, chiral)

    vecs = np.c_[prop_vecs[n:], evan_vecs[n:]]
    vecslmbdainv = np.c_[prop_vecs[:n], evan_vecs[:n]]

    # Prepare output for a single block
    wave_functions = real_space_data.wave_functions
    momenta = real_space_data.momenta
    velocities = real_space_data.velocities

    return (wave_functions, momenta, velocities, vecs, vecslmbdainv, v)


def transform_modes(modes_data, unitary=None, time_reversal=None,
                    particle_hole=None, chiral=None):
    """Transform the modes data for a given block of the Hamiltonian using a
    discrete symmetry (see arguments). The symmetry operator can also be
    specified as can also be identity, for the case when blocks are identical.

    Assume that modes_data has the form returned by compute_block_modes, i.e. a
    tuple (wave_functions, momenta, velocities, nmodes, vecs, vecslmbdainv, v)
    containing the block modes data.

    Assume that the symmetry operator is written in the proper basis (block
    basis, not full TB).

    """
    wave_functions, momenta, velocities, vecs, vecslmbdainv, v = modes_data

    # Copy to not overwrite modes from previous blocks
    wave_functions = wave_functions.copy()
    momenta = momenta.copy()
    velocities = velocities.copy()
    vecs = vecs.copy()
    vecslmbdainv = vecslmbdainv.copy()
    v = v.copy()

    nmodes = wave_functions.shape[1] // 2

    if unitary is not None:
        perm = np.arange(2*nmodes)
        conj = False
        flip_energy = False
    elif time_reversal is not None:
        unitary = time_reversal
        perm = np.arange(2*nmodes)[::-1]
        conj = True
        flip_energy = False
    elif particle_hole is not None:
        unitary = particle_hole
        perm = ((-1-np.arange(2*nmodes)) % nmodes +
                nmodes * (np.arange(2*nmodes) // nmodes))
        conj = True
        flip_energy = True
    elif chiral is not None:
        unitary = chiral
        perm = (np.arange(2*nmodes) % nmodes +
                nmodes * (np.arange(2*nmodes) < nmodes))
        conj = False
        flip_energy = True
    else:  # skip coverage
        raise ValueError("No relation between blocks was provided.")

    if conj:
        momenta *= -1
        vecs = vecs.conj()
        vecslmbdainv = vecslmbdainv.conj()
        wave_functions = wave_functions.conj()
        v = v.conj()

    if flip_energy:
        vecslmbdainv *= -1

    if flip_energy != conj:
        velocities *= -1

    wave_functions = unitary.dot(wave_functions)[:, perm]
    v = unitary.dot(v)
    vecs[:, :2*nmodes] = vecs[:, perm]
    vecslmbdainv[:, :2*nmodes] = vecslmbdainv[:, perm]
    velocities = velocities[perm]
    momenta = momenta[perm]
    return (wave_functions, momenta, velocities, vecs, vecslmbdainv, v)


def modes(h_cell, h_hop, tol=1e6, stabilization=None, *,
          discrete_symmetry=None, projectors=None, time_reversal=None,
          particle_hole=None, chiral=None):
    """Compute the eigendecomposition of a translation operator of a lead.

    Parameters
    ----------
    h_cell : numpy array, real or complex, shape (N,N) The unit cell
        Hamiltonian of the lead unit cell.
    h_hop : numpy array, real or complex, shape (N,M)
        The hopping matrix from a lead cell to the one on which self-energy
        has to be calculated (and any other hopping in the same direction).
    tol : float
        Numbers and differences are considered zero when they are smaller
        than `tol` times the machine precision.
    stabilization : sequence of 2 booleans or None
        Which steps of the eigenvalue prolem stabilization to perform. If the
        value is `None`, then Kwant chooses the fastest (and least stable)
        algorithm that is expected to be sufficient.  For any other value,
        Kwant forms the eigenvalue problem in the basis of the hopping singular
        values.  The first element set to `True` forces Kwant to add an
        anti-Hermitian term to the cell Hamiltonian before inverting. If it is
        set to `False`, the extra term will only be added if the cell
        Hamiltonian isn't invertible. The second element set to `True` forces
        Kwant to solve a generalized eigenvalue problem, and not to reduce it
        to the regular one.  If it is `False`, reduction to a regular problem
        is performed if possible.  Selecting the stabilization manually is
        mostly necessary for testing purposes.
    particle_hole : sparse or dense square matrix
        The unitary part of the particle-hole symmetry operator.
    time_reversal : sparse or dense square matrix
        The unitary part of the time-reversal symmetry operator.
    chiral : sparse or dense square matrix
        The chiral symmetry operator.
    projectors : an iterable of sparse or dense matrices
        Projectors that block diagonalize the Hamiltonian in accordance
        with a conservation law.

    Returns
    -------
    propagating : `~kwant.physics.PropagatingModes`
        Contains the array of the wave functions of propagating modes, their
        momenta, and their velocities. It can be used to identify the gauge in
        which the scattering problem is solved.
    stabilized : `~kwant.physics.StabilizedModes`
        A basis of propagating and evanescent modes used by the solvers.

    Notes
    -----
    The sorting of the propagating modes is fully described in the
    documentation for `~kwant.physics.PropagatingModes`.  In simple cases where
    bands do not cross, this ordering corresponds to "lowest modes first". In
    general, however, it is necessary to examine the band structure --
    something this function is not doing by design.

    Propagating modes with the same momentum are orthogonalized. All the
    propagating modes are normalized by current.

    `projectors`, `time_reversal`, `particle_hole`, and `chiral` affect the
    basis in which the scattering modes are expressed - see
    `~kwant.physics.DiscreteSymmetry` for details.

    This function uses the most stable and efficient algorithm for calculating
    the mode decomposition that the Kwant authors are aware about. Its details
    are to be published.
    """
    if discrete_symmetry is not None:
        projectors, time_reversal, particle_hole, chiral = discrete_symmetry
    n, m = h_hop.shape

    if h_cell.shape != (n, n):
        raise ValueError("Incompatible matrix sizes for h_cell and h_hop.")

    if not np.any(h_hop):
        wf = np.zeros((n, 0))
        v = np.zeros((m, 0))
        m = np.zeros((0, 0))
        vec = np.zeros((0,))
        return (PropagatingModes(wf, vec, vec), StabilizedModes(m, m, 0, v))

    ham = h_cell
    # Avoid the trouble of dealing with non-square hopping matrices.
    # TODO: How to avoid this while not doing a lot of book-keeping?
    hop = np.empty_like(ham, dtype=h_hop.dtype)
    hop[:, :m] = h_hop
    hop[:, m:] = 0

    # Provide default values to not deal with special cases.
    if projectors is None:
        projectors = [sp_identity(ham.shape[0], format='csr')]

    def to_dense_or_csr(matrix):
        if matrix is None:
            return csr_matrix(ham.shape)
        try:
            return matrix.tocsr()
        except AttributeError:
            return matrix

    symmetries = map(to_dense_or_csr,
                     (time_reversal, particle_hole, chiral))
    time_reversal, particle_hole, chiral = symmetries

    offsets = np.cumsum([0] + [projector.shape[1] for projector in projectors])
    indices = [slice(*i) for i in np.vstack([offsets[:-1], offsets[1:]]).T]
    projection_op = sp_hstack(projectors)

    def basis_change(a, antiunitary=False):
        b = projection_op
        # We need the extra transposes to ensure that sparse dot is used.
        if antiunitary:
            # b.T.conj() @ a @ b.conj()
            return (b.T.conj().dot((b.T.conj().dot(a)).T)).T
        else:
            # b.T.conj() @ a @ b
            return (b.T.dot((b.T.conj().dot(a)).T)).T

    # Conservation law basis
    ham_cons = basis_change(ham)
    hop_cons = basis_change(hop)
    trs = basis_change(time_reversal, True)
    phs = basis_change(particle_hole, True)
    sls = basis_change(chiral)

    # Check that the Hamiltonian has the conservation law
    block_modes = len(projectors) * [None]
    numbers_coords = combinations_with_replacement(enumerate(indices), 2)
    for (i, x), (j, y) in numbers_coords:
        h = ham_cons[x, y]
        t = hop_cons[x, y]
        # Symmetries that project from block x to block y
        symmetries = [symm[y, x] for symm in (trs, phs, sls)]
        symmetries = [(symm if nonzero_symm_projection(symm) else None) for
                      symm in symmetries]
        if i == j:
            if block_modes[i] is not None:
                continue
            # We did not compute this block yet.
            block_modes[i] = compute_block_modes(h, t, tol, stabilization,
                                                 *symmetries)
        else:
            if block_modes[j] is not None:
                # Modes in the block already computed.
                continue
            x, y = tuple(np.mgrid[x, x]), tuple(np.mgrid[y, y])

            if ham_cons[x].shape != ham_cons[y].shape:
                continue
            if (np.allclose(ham_cons[x], ham_cons[y]) and
                np.allclose(hop_cons[x], hop_cons[y])):
                unitary = sp_identity(h.shape[0])
            else:
                unitary = None
            if any(op is not None for op in symmetries + [unitary]):
                block_modes[j] = transform_modes(block_modes[i], unitary,
                                                 *symmetries)
    (wave_functions, momenta, velocities,
     vecs, vecslmbdainv, sqrt_hops) = zip(*block_modes)

    # Reorder by direction of propagation
    wave_functions = group_halves([(projector.dot(wf)).T for wf, projector in
                                   zip(wave_functions, projectors)]).T
    # Propagating modes object to return
    prop_modes = PropagatingModes(wave_functions, group_halves(velocities),
                                  group_halves(momenta))
    nmodes = [len(v) // 2 for v in velocities]
    # Store the number of modes per block as an attribute.
    # nmodes[i] is the number of left or right moving modes in block i.
    # In the module that makes leads with conservation laws, this is necessary
    # to keep track of the block structure of the scattering matrix.
    prop_modes.block_nmodes = nmodes

    parts = zip(*([v[:, :n], v[:, n:2*n], v[:, 2*n:]]
                  for n, v in zip(nmodes, vecs)))
    vecs = np.hstack([block_diag(*part) for part in parts])

    parts = zip(*([v[:, :n], v[:, n:2*n], v[:, 2*n:]]
                  for n, v in zip(nmodes, vecslmbdainv)))
    vecslmbdainv = np.hstack([block_diag(*part) for part in parts])

    sqrt_hops = np.hstack([projector.dot(hop) for projector, hop in
                             zip(projectors, sqrt_hops)])[:h_hop.shape[1]]

    stab_modes = StabilizedModes(vecs, vecslmbdainv, sum(nmodes), sqrt_hops)

    return prop_modes, stab_modes


def selfenergy(h_cell, h_hop, tol=1e6):
    """
    Compute the self-energy generated by the lead.

    Parameters
    ----------
    h_cell : numpy array, real or complex, shape (N,N) The unit cell Hamiltonian
        of the lead unit cell.
    h_hop : numpy array, real or complex, shape (N,M)
        The hopping matrix from a lead cell to the one on which self-energy
        has to be calculated (and any other hopping in the same direction).
    tol : float
        Numbers are considered zero when they are smaller than `tol` times
        the machine precision.

    Returns
    -------
    Sigma : numpy array, real or complex, shape (M,M)
        The computed self-energy. Note that even if `h_cell` and `h_hop` are
        both real, `Sigma` will typically be complex. (More precisely, if there
        is a propagating mode, `Sigma` will definitely be complex.)

    Notes
    -----
    For simplicity this function internally calculates the modes first.
    This may cause a small slowdown, and can be improved if necessary.
    """
    stabilized = modes(h_cell, h_hop, tol)[1]
    return stabilized.selfenergy()


def square_selfenergy(width, hopping, fermi_energy):
    """
    Calculate analytically the self energy for a square lattice.

    The lattice is assumed to have a single orbital per site and
    nearest-neighbor hopping.

    Parameters
    ----------
    width : integer
        width of the lattice
    """

    # Following appendix C of M. Wimmer's diploma thesis:
    # http://www.physik.uni-regensburg.de/forschung/\
    # richter/richter/media/research/publications2004/wimmer-Diplomarbeit.pdf

    # p labels transversal modes.  i and j label the sites of a cell.

    # Precalculate the transverse wave function.
    psi_p_i = np.empty((width, width))
    factor = pi / (width + 1)
    prefactor = sqrt(2 / (width + 1))
    for p in range(width):
        for i in range(width):
            psi_p_i[p, i] = prefactor * sin(factor * (p + 1) * (i + 1))

    # Precalculate the integrals of the longitudinal wave functions.
    def f(q):
        if abs(q) <= 2:
            return q/2 - 1j * sqrt(1 - (q / 2) ** 2)
        else:
            return q/2 - copysign(sqrt((q / 2) ** 2 - 1), q)
    f_p = np.empty((width,), dtype=complex)
    for p in range(width):
        e = 2 * hopping * (1 - cos(factor * (p + 1)))
        q = (fermi_energy - e) / hopping - 2
        f_p[p] = f(q)

    # Put everything together into the self energy and return it.
    result = np.empty((width, width), dtype=complex)
    for i in range(width):
        for j in range(width):
            result[i, j] = hopping * (
                psi_p_i[:, i] * psi_p_i[:, j].conj() * f_p).sum()
    return result



# Copyright 2011-2018 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import collections
import inspect
import cmath
import warnings

import tinyarray as ta
import numpy as np
import scipy.linalg
import scipy.spatial

from . import builder, system, plotter
from .linalg import lll
from .builder import herm_conj, HermConjOfFunc
from .lattice import TranslationalSymmetry
from ._common import get_parameters


__all__ = ['wraparound', 'plot_2d_bands']


def _hashable(obj):
    return isinstance(obj, collections.abc.Hashable)


def _memoize(f):
    """Decorator to memoize a function that works even with unhashable args.

    This decorator will even work with functions whose args are not hashable.
    The cache key is made up by the hashable arguments and the ids of the
    non-hashable args.  It is up to the user to make sure that non-hashable
    args do not change during the lifetime of the decorator.

    This decorator will keep reevaluating functions that return None.
    """
    def lookup(*args):
        key = tuple(arg if _hashable(arg) else id(arg) for arg in args)
        result = cache.get(key)
        if result is None:
            cache[key] = result = f(*args)
        return result
    cache = {}
    return lookup


def _set_signature(func, params):
    """Set the signature of 'func'.

    Parameters
    ----------
    func : callable
    params: sequence of str
        Parameter names to put in the signature. These will be added as
        'POSITIONAL_ONLY' type parameters.
    """
    params = [inspect.Parameter(name, inspect.Parameter.POSITIONAL_ONLY)
              for name in params]
    func.__signature__ = inspect.Signature(params)


## This wrapper is needed so that finalized systems that
## have been wrapped can be queried for their symmetry, which
## is needed for Brillouin zone calculations (plotting).

class WrappedBuilder(builder.Builder):

    def finalized(self):
        ret = super().finalized()
        ret._momentum_names = self._momentum_names
        ret._wrapped_symmetry = self._wrapped_symmetry
        return ret


def wraparound(builder, keep=None, *, coordinate_names='xyz'):
    """Replace translational symmetries by momentum parameters.

    A new Builder instance is returned.  By default, each symmetry is replaced
    by one scalar momentum parameter that is appended to the already existing
    arguments of the system.  Optionally, one symmetry may be kept by using the
    `keep` argument. The momentum parameters will have names like 'k_n' where
    the 'n' are specified by 'coordinate_names'.

    Parameters
    ----------
    builder : `~kwant.builder.Builder`
    keep : int, optional
        Which (if any) translational symmetry to keep.
    coordinate_names : sequence of strings, default: 'xyz'
        The names of the coordinates along the symmetry
        directions of 'builder'.

    Notes
    -----
    Wraparound is stop-gap functionality until Kwant 2.x which will include
    support for higher-dimension translational symmetry in the low-level system
    format. It will be deprecated in the 2.0 release of Kwant.
    """

    @_memoize
    def bind_site(val):
        def f(*args):
            a, *args = args
            return val(a, *args[:mnp])

        assert callable(val)
        _set_signature(f, get_parameters(val) + momenta)
        return f

    @_memoize
    def bind_hopping_as_site(elem, val):
        def f(*args):
            a, *args = args
            phase = cmath.exp(1j * ta.dot(elem, args[mnp:]))
            v = val(a, sym.act(elem, a), *args[:mnp]) if callable(val) else val
            pv = phase * v
            return pv + herm_conj(pv)

        params = ('_site0',)
        if callable(val):
            params += get_parameters(val)[2:] # cut off both site parameters
        _set_signature(f, params + momenta)
        return f

    @_memoize
    def bind_hopping(elem, val):
        def f(*args):
            a, b, *args = args
            phase = cmath.exp(1j * ta.dot(elem, args[mnp:]))
            v = val(a, sym.act(elem, b), *args[:mnp]) if callable(val) else val
            return phase * v

        params = ('_site0', '_site1')
        if callable(val):
            params += get_parameters(val)[2:] # cut off site parameters
        _set_signature(f, params + momenta)
        return f

    @_memoize
    def bind_sum(num_sites, *vals):
        """Construct joint signature for all 'vals'."""

        def f(*in_args):
            acc = 0
            for val, selection in val_selection_pairs:
                if selection:   # Otherwise: reuse previous out_args.
                    out_args = tuple(in_args[i] for i in selection)
                if callable(val):
                    acc = acc + val(*out_args)
                else:
                    acc = acc + val
            return acc

        params = collections.OrderedDict()

        # Add the leading one or two 'site' parameters.
        site_params = ['_site{}'.format(i) for i in range(num_sites)]
        for name in site_params:
            params[name] = inspect.Parameter(
                name, inspect.Parameter.POSITIONAL_ONLY)

        # Add all the other parameters (except for the momenta).  Setup the
        # 'selections'.
        selections = []
        for val in vals:
            if not callable(val):
                selections.append(())
                continue
            val_params = get_parameters(val)[num_sites:]
            assert val_params[mnp:] == momenta
            val_params = val_params[:mnp]
            selections.append((*site_params, *val_params))
            for p in val_params:
                # Skip parameters that exist in previously added functions.
                if p in params:
                    continue
                params[p] = inspect.Parameter(
                    p, inspect.Parameter.POSITIONAL_ONLY)

        # Sort values such that ones with the same arguments are bunched.
        # Prepare 'val_selection_pairs' that is used in the function 'f' above.
        params_keys = list(params.keys())
        val_selection_pairs = []
        prev_selection = None
        argsort = sorted(range(len(selections)), key=selections.__getitem__)
        momenta_sel = tuple(range(mnp, 0, 1))
        for i in argsort:
            selection = selections[i]
            if selection and selection != prev_selection:
                prev_selection = selection = tuple(
                    params_keys.index(s) for s in selection) + momenta_sel
            else:
                selection = ()
            val_selection_pairs.append((vals[i], selection))

        # Finally, add the momenta.
        for k in momenta:
            params[k] = inspect.Parameter(
                k, inspect.Parameter.POSITIONAL_ONLY)

        f.__signature__ = inspect.Signature(params.values())
        return f

    try:
        momenta = ['k_{}'.format(coordinate_names[i])
                   for i in range(len(builder.symmetry.periods))]
    except IndexError:
        raise ValueError("All symmetry directions must have a name specified "
                         "in coordinate_names")

    if keep is None:
        ret = WrappedBuilder()
        sym = builder.symmetry
    else:
        periods = list(builder.symmetry.periods)
        ret = WrappedBuilder(TranslationalSymmetry(periods.pop(keep)))
        sym = TranslationalSymmetry(*periods)
        momenta.pop(keep)
    momenta = tuple(momenta)
    mnp = -len(momenta)         # Used by the bound functions above.

    # Store the names of the momentum parameters and the symmetry of the
    # old Builder (this will be needed for band structure plotting)
    ret._momentum_names = momenta
    ret._wrapped_symmetry = builder.symmetry

    # Wrapped around system retains conservation law and chiral symmetry.
    # We use 'bind_site' to add the momenta arguments if required.
    cons = builder.conservation_law
    ret.conservation_law = bind_site(cons) if callable(cons) else cons
    chiral = builder.chiral
    ret.chiral = bind_site(chiral) if callable(chiral) else chiral

    if builder.particle_hole is not None or builder.time_reversal is not None:
        warnings.warn('`particle_hole` and `time_reversal` symmetries are set '
                      'on the input builder. However they are ignored for the '
                      'wrapped system, since Kwant lacks a way to express the '
                      'existence (or not) of a symmetry at k != 0.',
                      RuntimeWarning, stacklevel=2)
    ret.particle_hole = None
    ret.time_reversal = None

    sites = {}
    hops = collections.defaultdict(list)

    # Store lists of values, so that multiple values can be assigned to the
    # same site or hopping.
    for site, val in builder.site_value_pairs():
        # Every 'site' is in the FD of the original symmetry.
        # Move the sites to the FD of the remaining symmetry, this guarantees that
        # every site in the new system is an image of an original FD site translated
        # purely by the remaining symmetry.
        sites[ret.symmetry.to_fd(site)] = [bind_site(val) if callable(val) else val]

    for hop, val in builder.hopping_value_pairs():
        a, b = hop
        # 'a' is in the FD of original symmetry.
        # Get translation from FD of original symmetry to 'b',
        # this is different from 'b_dom = sym.which(b)'.
        b_dom = builder.symmetry.which(b)
        # Throw away part that is in the remaining translation direction, so we get
        # an element of 'sym' which is being wrapped
        b_dom = ta.array([t for i, t in enumerate(b_dom) if i != keep])
        # Pull back using the remainder, which is purely in the wrapped directions.
        # This guarantees that 'b_wa' is an image of an original FD site translated
        # purely by the remaining symmetry.
        b_wa = sym.act(-b_dom, b)
        # Move the hopping to the FD of the remaining symmetry
        a, b_wa = ret.symmetry.to_fd(a, b_wa)

        if a == b_wa:
            # The hopping gets wrapped-around into an onsite Hamiltonian.
            # Since site `a` already exists in the system, we can simply append.
            sites[a].append(bind_hopping_as_site(b_dom, val))
        else:
            # The hopping remains a hopping.
            if any(b_dom) or callable(val):
                # The hopping got wrapped-around or is a function.
                val = bind_hopping(b_dom, val)

            # Make sure that there is only one entry for each hopping
            # pointing in one direction, modulo the remaining translations.
            b_wa_r, a_r = ret.symmetry.to_fd(b_wa, a)
            if (b_wa_r, a_r) in hops:
                assert (a, b_wa) not in hops
                if callable(val):
                    assert not isinstance(val, HermConjOfFunc)
                    val = HermConjOfFunc(val)
                else:
                    val = herm_conj(val)

                hops[b_wa_r, a_r].append(val)
            else:
                hops[a, b_wa].append(val)

    # Copy stuff into result builder, converting lists of more than one element
    # into summing functions.
    for site, vals in sites.items():
        ret[site] = vals[0] if len(vals) == 1 else bind_sum(1, *vals)

    for hop, vals in hops.items():
        ret[hop] = vals[0] if len(vals) == 1 else bind_sum(2, *vals)

    return ret


def plot_2d_bands(syst, k_x=31, k_y=31, params=None,
                  mask_brillouin_zone=False, extend_bbox=0, file=None,
                  show=True, dpi=None, fig_size=None, ax=None):
    """Plot 2D band structure of a wrapped around system.

    This function is primarily useful for systems that have translational
    symmetry vectors that are non-orthogonal (e.g. graphene). This function
    will properly plot the band structure in an orthonormal basis in k-space,
    as opposed to in the basis of reciprocal lattice vectors (which would
    produce a "skewed" Brillouin zone).

    If your system has orthogonal lattice vectors, you are probably better
    off using `kwant.plotter.spectrum`.

    Parameters
    ----------
    syst : `kwant.system.FiniteSystem`
        A 2D system that was finalized from a Builder produced by
        `kwant.wraparound.wraparound`.  Note that this *must* be a finite
        system; so `kwant.wraparound.wraparound` should have been called with
        ``keep=None``.
    k_x, k_y : int or sequence of float, default: 31
        Either a number of sampling points, or a sequence of points at which
        the band structure is to be evaluated, in units of inverse length.
    params : dict, optional
        Dictionary of parameter names and their values, not including the
        momentum parameters.
    mask_brillouin_zone : bool, default: False
        If True, then the band structure will only be plotted over the first
        Brillouin zone. By default the band structure is plotted over a
        rectangular bounding box that contains the Brillouin zone.
    extend_bbox : float, default: 0
        Amount by which to extend the region over which the band structure is
        plotted, expressed as a proportion of the Brillouin zone bounding box
        length. i.e. ``extend_bbox=0.1`` will extend the region by 10% (in all
        directions).
    file : string or file object, optional
        The output file.  If None, output will be shown instead.
    show : bool, default: False
        Whether ``matplotlib.pyplot.show()`` is to be called, and the output is
        to be shown immediately.  Defaults to `True`.
    dpi : float, optional
        Number of pixels per inch.  If not set the ``matplotlib`` default is
        used.
    fig_size : tuple, optional
        Figure size `(width, height)` in inches.  If not set, the default
        ``matplotlib`` value is used.
    ax : ``matplotlib.axes.Axes`` instance, optional
        If `ax` is not `None`, no new figure is created, but the plot is done
        within the existing Axes `ax`. in this case, `file`, `show`, `dpi`
        and `fig_size` are ignored.

    Returns
    -------
    fig : matplotlib figure
        A figure with the output if `ax` is not set, else None.

    Notes
    -----
    This function produces plots where the units of momentum are inverse
    length. This is contrary to `kwant.plotter.bands`, where the units
    of momentum are inverse lattice constant.

    If the lattice vectors for the symmetry of ``syst`` are not orthogonal,
    then part of the plotted band structure will be outside the first Brillouin
    zone (inside the bounding box of the brillouin zone). Setting
    ``mask_brillouin_zone=True`` will cause the plot to be truncated outside of
    the first Brillouin zone.

    See Also
    --------
    kwant.plotter.spectrum
    """
    if not hasattr(syst, '_wrapped_symmetry'):
        raise TypeError("Expecting a system that was produced by "
                        "'kwant.wraparound.wraparound'.")
    if not isinstance(syst, system.FiniteSystem):
        msg = ("All symmetry directions must be wrapped around: specify "
               "'keep=None' when calling 'kwant.wraparound.wraparound'.")
        raise TypeError(msg)

    params = params or {}
    lat_ndim, space_ndim = syst._wrapped_symmetry.periods.shape

    if lat_ndim != 2:
        raise ValueError("Expected a system with a 2D translational symmetry.")
    if space_ndim != lat_ndim:
        raise ValueError("Lattice dimension must equal realspace dimension.")

    # columns of B are lattice vectors
    B = np.array(syst._wrapped_symmetry.periods).T
    # columns of A are reciprocal lattice vectors
    A = np.linalg.pinv(B).T

    ## calculate the bounding box for the 1st Brillouin zone

    # Get lattice points that neighbor the origin, in basis of lattice vectors
    reduced_vecs, transf = lll.lll(A.T)
    neighbors = ta.dot(lll.voronoi(reduced_vecs), transf)
    # Add the origin to these points.
    klat_points = np.concatenate(([[0] * lat_ndim], neighbors))
    # Transform to cartesian coordinates and rescale.
    # Will be used in 'outside_bz' function, later on.
    klat_points = 2 * np.pi * np.dot(klat_points, A.T)
    # Calculate the Voronoi cell vertices
    vor = scipy.spatial.Voronoi(klat_points)
    around_origin = vor.point_region[0]
    bz_vertices = vor.vertices[vor.regions[around_origin]]
    # extract bounding box
    k_max = np.max(np.abs(bz_vertices), axis=0)

    ## build grid along each axis, if needed
    ks = []
    for k, km in zip((k_x, k_y), k_max):
        k = np.array(k)
        if not k.shape:
            if extend_bbox:
                km += km * extend_bbox
            k = np.linspace(-km, km, k)
        ks.append(k)

    # TODO: It is very inefficient to call 'momentum_to_lattice' once for
    #       each point (for trivial Hamiltonians 60% of the time is spent
    #       doing this). We should instead transform the whole grid in one call.

    def momentum_to_lattice(k):
        k, residuals = scipy.linalg.lstsq(A, k)[:2]
        if np.any(abs(residuals) > 1e-7):
            raise RuntimeError("Requested momentum doesn't correspond"
                               " to any lattice momentum.")
        return k

    def ham(k_x, k_y=None, **params):
        # transform into the basis of reciprocal lattice vectors
        k = momentum_to_lattice([k_x] if k_y is None else [k_x, k_y])
        p = dict(zip(syst._momentum_names, k), **params)
        return syst.hamiltonian_submatrix(params=p, sparse=False)

    def outside_bz(k_x, k_y, **_):
        dm = scipy.spatial.distance_matrix(klat_points, [[k_x, k_y]])
        return np.argmin(dm) != 0  # is origin no closest 'klat_point' to 'k'?

    fig = plotter.spectrum(ham,
                           x=('k_x', ks[0]),
                           y=('k_y', ks[1]) if lat_ndim == 2 else None,
                           params=params,
                           mask=(outside_bz if mask_brillouin_zone else None),
                           file=file, show=show, dpi=dpi,
                           fig_size=fig_size, ax=ax)
    return fig



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


import numpy as np
from kwant.linalg import lll
from kwant._common import ensure_rng

def test_lll():
    rng = ensure_rng(1)
    for i in range(50):
        x = rng.randint(4) + 1
        mat = rng.randn(x, x + rng.randint(2))
        c = 1.34 + .5 * rng.random_sample()
        reduced_mat, coefs = lll.lll(mat)
        assert lll.is_c_reduced(reduced_mat, c)
        assert np.allclose(np.dot(mat.T, coefs), reduced_mat.T)


def test_cvp():
    rng = ensure_rng(0)
    for i in range(1, 5):
        for j in range(i, 5):
            mat = rng.randn(i, j)
            mat = lll.lll(mat)[0]
            for k in range(4):
                point = 50 * rng.randn(j)
                assert np.array_equal(lll.cvp(point, mat, 10)[:3],
                                      lll.cvp(point, mat, 3))

    # Test equidistant vectors
    # Cubic lattice
    basis = np.eye(3)
    vec = np.zeros((3))
    assert len(lll.cvp(vec, basis, n=2, group_by_length=True)) == 7
    assert len(lll.cvp(vec, basis, n=3, group_by_length=True)) == 19
    vec = 0.5 * np.array([1, 1, 1])
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 8
    vec = 0.5 * np.array([1, 1, 0])
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 4
    vec = 0.5 * np.array([1, 0, 0])
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 2
    # Square lattice with offset
    offset = np.array([0, 0, 1])
    basis = np.eye(3)[:2]
    vec = np.zeros((3)) + rng.rand() * offset
    assert len(lll.cvp(vec, basis, n=2, group_by_length=True)) == 5
    assert len(lll.cvp(vec, basis, n=3, group_by_length=True)) == 9
    vec = 0.5 * np.array([1, 1, 0]) + rng.rand() * offset
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 4
    vec = 0.5 * np.array([1, 0, 0]) + rng.rand() * offset
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 2
    # Hexagonal lattice
    basis = np.array([[1, 0], [-0.5, 0.5 * np.sqrt(3)]])
    vec = np.zeros((2))
    assert len(lll.cvp(vec, basis, n=2, group_by_length=True)) == 7
    assert len(lll.cvp(vec, basis, n=3, group_by_length=True)) == 13
    vec = np.array([0.5, 0.5 / np.sqrt(3)])
    assert len(lll.cvp(vec, basis, group_by_length=True)) == 3
    assert len(lll.cvp(vec, basis, n=2, group_by_length=True)) == 6
    assert len(lll.cvp(vec, basis, n=3, group_by_length=True)) == 12



from collections import namedtuple, Counter
import warnings
from math import sqrt
import numpy as np
import pytest

import kwant
from ... import lattice
from ...builder import HoppingKind, Builder, NoSymmetry, Site
from .. import gauge


## Utilities

# TODO: remove in favour of 'scipy.stats.special_ortho_group' once
#       we depend on scipy 0.18
class special_ortho_group_gen:

    def rvs(self, dim):
        H = np.eye(dim)
        D = np.empty((dim,))
        for n in range(dim-1):
            x = np.random.normal(size=(dim-n,))
            D[n] = np.sign(x[0]) if x[0] != 0 else 1
            x[0] += D[n]*np.sqrt((x*x).sum())
            # Householder transformation
            Hx = (np.eye(dim-n)
                  - 2.*np.outer(x, x)/(x*x).sum())
            mat = np.eye(dim)
            mat[n:, n:] = Hx
            H = np.dot(H, mat)
        D[-1] = (-1)**(dim-1)*D[:-1].prod()
        # Equivalent to np.dot(np.diag(D), H) but faster, apparently
        H = (D*H.T).T
        return H

special_ortho_group = special_ortho_group_gen()


square_lattice = lattice.square(norbs=1, name='square')
honeycomb_lattice = lattice.honeycomb(norbs=1, name='honeycomb')
cubic_lattice = lattice.cubic(norbs=1, name='cubic')

def rectangle(W, L):
    return (
        lambda s: 0 <= s.pos[0] < L and 0 <= s.pos[1] < W,
        (L/2, W/2)
    )


def ring(r_inner, r_outer):
    return (
        lambda s: r_inner <= np.linalg.norm(s.pos) <= r_outer,
        ((r_inner + r_outer) / 2, 0)
    )

def wedge(W):
    return (
        lambda s: (0 <= s.pos[0] < W) and (0 <= s.pos[1] <= s.pos[0]),
        (0, 0)
    )


def half_ring(r_inner, r_outer):
    in_ring, _ = ring(r_inner, r_outer)
    return (
        lambda s: s.pos[0] <= 0 and in_ring(s),
        (-(r_inner + r_outer) / 2, 0)
    )


def cuboid(a, b, c):
    return (
        lambda s: 0 <= s.pos[0] < a and 0 <= s.pos[1] < b and 0 <= s.pos[2] < c,
        (a/2, b/2, c/2)
    )

def hypercube(dim, W):
    return (
        lambda s: all(0 <= x < W for x in s.pos),
        (W / 2,) * dim
    )


def circle(r):
    return (
        lambda s: np.linalg.norm(s.pos) < r,
        (0, 0)
    )

def ball(dim, r):
    return (
        lambda s: np.linalg.norm(s.pos) < r,
        (0,) * dim
    )


def model(lat, neighbors):
    syst = Builder(lattice.TranslationalSymmetry(*lat.prim_vecs))
    if hasattr(lat, 'sublattices'):
        for l in lat.sublattices:
            zv = (0,) * len(l.prim_vecs)
            syst[l(*zv)] = None
    else:
        zv = (0,) * len(l.prim_vecs)
        syst[lat(*zv)] = None
    for r in range(neighbors):
        syst[lat.neighbors(r + 1)] = None
    return syst


def check_loop_kind(loop_kind):
    (_, first_fam_a, prev_fam_b), *rest = loop_kind
    for (_, fam_a, fam_b) in rest:
        if prev_fam_b != fam_a:
            raise ValueError('Invalid loop kind: does not close')
        prev_fam_b = fam_b
    # loop closes
    net_delta = np.sum([hk.delta for hk in loop_kind])
    if first_fam_a != fam_b or np.any(net_delta != 0):
        raise ValueError('Invalid loop kind: does not close')


def available_loops(syst, loop_kind):

    def maybe_loop(site):
        loop = [site]
        a = site
        for delta, family_a, family_b in loop_kind:
            b = Site(family_b, a.tag + delta, True)
            if family_a != a.family or (a, b) not in syst:
                return None
            loop.append(b)
            a = b
        return loop

    check_loop_kind(loop_kind)

    return list(filter(None, map(maybe_loop, syst.sites())))


def loop_to_links(loop):
    return list(zip(loop, loop[1:]))


def no_symmetry(lat, neighbors):
    return NoSymmetry()


def translational_symmetry(lat, neighbors):
    return lattice.TranslationalSymmetry(int((neighbors + 1)/2) * lat.prim_vecs[0])


## Tests

# Tests that phase around a loop is equal to the flux through the loop.
# First we define the loops that we want to test, for various lattices.
# If a system does not support a particular kind of loop, they will simply
# not be generated.

Loop = namedtuple('Loop', ('path', 'flux'))

square_loops = [([HoppingKind(d, square_lattice) for d in l.path], l.flux)
        for l in [
    # 1st nearest neighbors
    Loop(path=[(1, 0), (0, 1), (-1, 0), (0, -1)], flux=1),
    # 2nd nearest neighbors
    Loop(path=[(1, 0), (0, 1), (-1, -1)], flux=0.5),
    Loop(path=[(1, 0), (-1, 1), (0, -1)], flux=0.5),
    # 3rd nearest neighbors
    Loop(path=[(2, 0), (0, 1), (-2, 0), (0, -1)], flux=2),
    Loop(path=[(2, 0), (-1, 1), (-1, 0), (0, -1)], flux=1.5),
]]

a, b = honeycomb_lattice.sublattices
honeycomb_loops = [([HoppingKind(d, a, b) for *d, a, b in l.path], l.flux)
        for l in [
    # 1st nearest neighbors
    Loop(path=[(0, 0, a, b), (-1, 1, b, a), (0, -1, a, b), (0, 0, b, a),
               (1, -1, a, b), (0, 1, b, a)],
         flux=sqrt(3)/2),
    # 2nd nearest neighbors
    Loop(path=[(-1, 1, a, a), (0, -1, a, a), (1, 0, a, a)],
         flux=sqrt(3)/4),
    Loop(path=[(-1, 0, b, b), (1, -1, b, b), (0, 1, b, b)],
         flux=sqrt(3)/4),
]]

cubic_loops = [([HoppingKind(d, cubic_lattice) for d in l.path], l.flux)
        for l in [
    # 1st nearest neighbors
    Loop(path=[(1, 0, 0), (0, 1, 0), (-1, 0, 0), (0, -1, 0)], flux=1),
    Loop(path=[(0, 1, 0), (0, 0, 1), (0, -1, 0), (0, 0, -1)], flux=0),
    Loop(path=[(1, 0, 0), (0, 0, 1), (-1, 0, 0), (0, 0, -1)], flux=0),
    # 2nd nearest neighbors
    Loop(path=[(1, 0, 0), (-1, 1, 0), (0, -1, 0)], flux=0.5),
    Loop(path=[(1, 0, 0), (0, 1, 0), (-1, -1, 0)], flux=0.5),
    Loop(path=[(1, 0, 0), (-1, 0, 1), (0, 0, -1)], flux=0),
    Loop(path=[(1, 0, 0), (0, 0, 1), (-1, 0, -1)], flux=0),
    Loop(path=[(0, 1, 0), (0, -1, 1), (0, 0, -1)], flux=0),
    Loop(path=[(0, 1, 0), (0, 0, 1), (0, -1, -1)], flux=0),
    # 3rd nearest neighbors
    Loop(path=[(1, 1, 1), (0, 0, -1), (-1, -1, 0)], flux=0),
    Loop(path=[(1, 1, 1), (-1, 0, -1), (0, -1, 0)], flux=0.5),
]]

square = (square_lattice, square_loops)
honeycomb = (honeycomb_lattice, honeycomb_loops)
cubic = (cubic_lattice, cubic_loops)


def _test_phase_loops(syst, phases, loops):
    for loop_kind, loop_flux in loops:
        for loop in available_loops(syst, loop_kind):
            loop_phase = np.prod([phases(a, b) for a, b in loop_to_links(loop)])
            expected_loop_phase = np.exp(1j * np.pi * loop_flux)
            assert np.isclose(loop_phase, expected_loop_phase)


@pytest.mark.parametrize("neighbors", [1, 2, 3])
@pytest.mark.parametrize("symmetry", [no_symmetry, translational_symmetry],
                         ids=['finite', 'infinite'])
@pytest.mark.parametrize("lattice, loops", [square, honeycomb, cubic],
                         ids=['square', 'honeycomb', 'cubic'])
def test_phases(lattice, neighbors, symmetry, loops):
    """Check that the phases around common loops are equal to the flux, for
    finite and infinite systems with uniform magnetic field.
    """
    W = 4
    dim = len(lattice.prim_vecs)
    field = np.array([0, 0, 1]) if dim == 3 else 1

    syst = Builder(symmetry(lattice, neighbors))
    syst.fill(model(lattice, neighbors), *hypercube(dim, W))

    this_gauge = gauge.magnetic_gauge(syst.finalized())
    phases = this_gauge(field)

    _test_phase_loops(syst, phases, loops)


@pytest.mark.parametrize("neighbors", [1, 2, 3])
@pytest.mark.parametrize("lat, loops", [square, honeycomb],
                         ids=['square', 'honeycomb'])
def test_phases_composite(neighbors, lat, loops):
    """Check that the phases around common loops are equal to the flux, for
    composite systems with uniform magnetic field.
    """
    W = 4
    dim = len(lat.prim_vecs)
    field = np.array([0, 0, 1]) if dim == 3 else 1

    lead = Builder(lattice.TranslationalSymmetry(-lat.prim_vecs[0]))
    lead.fill(model(lat, neighbors), *hypercube(dim, W))

    # Case where extra sites are added and fields are same in
    # scattering region and lead.
    syst = Builder()
    syst.fill(model(lat, neighbors), *ball(dim, W + 1))
    extra_sites = syst.attach_lead(lead)
    assert extra_sites  # make sure we're testing the edge case with added sites

    this_gauge = gauge.magnetic_gauge(syst.finalized())
    # same field in scattering region and lead
    phases, lead_phases = this_gauge(field, field)

    # When extra sites are added to the central region we need to select
    # the correct phase function.
    def combined_phases(a, b):
        if a in extra_sites or b in extra_sites:
            return lead_phases(a, b)
        else:
            return phases(a, b)

    _test_phase_loops(syst, combined_phases, loops)
    _test_phase_loops(lead, lead_phases, loops)


@pytest.mark.parametrize("neighbors", [1, 2])
def test_overlapping_interfaces(neighbors):
    """Test composite systems with overlapping lead interfaces."""

    lat = square_lattice

    def _make_syst(edge, W=5):

        syst = Builder()
        syst.fill(model(lat, neighbors), *rectangle(W, W))

        leadx = Builder(lattice.TranslationalSymmetry((-1, 0)))
        leadx[(lat(0, j) for j in range(edge, W - edge))] = None
        for n in range(1, neighbors + 1):
            leadx[lat.neighbors(n)] = None

        leady = Builder(lattice.TranslationalSymmetry((0, -1)))
        leady[(lat(i, 0) for i in range(edge, W - edge))] = None
        for n in range(1, neighbors + 1):
            leady[lat.neighbors(n)] = None

        assert not syst.attach_lead(leadx)  # sanity check; no sites added
        assert not syst.attach_lead(leady)  # sanity check; no sites added

        return syst, leadx, leady

    # edge == 0: lead interfaces overlap
    # edge == 1: lead interfaces partition scattering region
    #            into 2 disconnected components
    for edge in (0, 1):
        syst, leadx, leady = _make_syst(edge)
        this_gauge = gauge.magnetic_gauge(syst.finalized())
        phases, leadx_phases, leady_phases = this_gauge(1, 1, 1)
        _test_phase_loops(syst, phases, square_loops)
        _test_phase_loops(leadx, leadx_phases, square_loops)
        _test_phase_loops(leady, leady_phases, square_loops)


def _make_square_syst(sym, neighbors=1):
    lat = square_lattice
    syst = Builder(sym)
    syst[(lat(i, j) for i in (0, 1) for j in (0, 1))] = None
    for n in range(1, neighbors + 1):
        syst[lat.neighbors(n)] = None
    return syst


def test_unfixable_gauge():
    """Check error is raised when we cannot fix the gauge."""

    leadx = _make_square_syst(lattice.TranslationalSymmetry((-1, 0)))
    leady = _make_square_syst(lattice.TranslationalSymmetry((0, -1)))

    # 1x2 line with 2 leads
    syst = _make_square_syst(NoSymmetry())
    del syst[[square_lattice(1, 0), square_lattice(1, 1)]]
    syst.attach_lead(leadx)
    syst.attach_lead(leadx.reversed())
    with pytest.raises(ValueError):
        gauge.magnetic_gauge(syst.finalized())

    # 2x2 square with leads attached from all 4 sides,
    # and nearest neighbor hoppings
    syst = _make_square_syst(NoSymmetry())
    # Until we add the last lead we have enough gauge freedom
    # to specify independent fields in the scattering region
    # and each of the leads. We check that no extra sites are
    # added as a sanity check.
    assert not syst.attach_lead(leadx)
    gauge.magnetic_gauge(syst.finalized())
    assert not syst.attach_lead(leady)
    gauge.magnetic_gauge(syst.finalized())
    assert not syst.attach_lead(leadx.reversed())
    gauge.magnetic_gauge(syst.finalized())
    # Adding the last lead removes our gauge freedom.
    assert not syst.attach_lead(leady.reversed())
    with pytest.raises(ValueError):
        gauge.magnetic_gauge(syst.finalized())

    # 2x2 square with 2 leads, but 4rd nearest neighbor hoppings
    syst = _make_square_syst(NoSymmetry())
    del syst[(square_lattice(1, 0), square_lattice(1, 1))]
    leadx = _make_square_syst(lattice.TranslationalSymmetry((-1, 0)))
    leadx[square_lattice.neighbors(4)] = None
    for lead in (leadx, leadx.reversed()):
        syst.attach_lead(lead)

    with pytest.raises(ValueError):
        gauge.magnetic_gauge(syst.finalized())


def _test_disconnected(syst):
    with pytest.raises(ValueError) as excinfo:
        gauge.magnetic_gauge(syst.finalized())
        assert 'unit cell not connected' in str(excinfo.value)

def test_invalid_lead():
    """Check error is raised when a lead unit cell is not connected
       within the unit cell itself.

       In order for the algorithm to work we need to be able to close
       loops within the lead. However we only add a single lead unit
       cell, so not all paths can be closed, even if the lead is
       connected.
    """
    lat = square_lattice

    lead = _make_square_syst(lattice.TranslationalSymmetry((-1, 0)),
                             neighbors=0)
    # Truly disconnected system
    # Ignore warnings to suppress Kwant's complaint about disconnected lead
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        _test_disconnected(lead)

    # 2 disconnected chains (diagonal)
    lead[(lat(0, 0), lat(1, 1))] = None
    lead[(lat(0, 1), lat(1, 0))] = None
    _test_disconnected(lead)

    lead = _make_square_syst(lattice.TranslationalSymmetry((-1, 0)),
                             neighbors=0)
    # 2 disconnected chains (horizontal)
    lead[(lat(0, 0), lat(1, 0))] = None
    lead[(lat(0, 1), lat(1, 1))] = None
    _test_disconnected(lead)

    # System has loops, but need 3 unit cells
    # to express them.
    lead[(lat(0, 0), lat(1, 1))] = None
    lead[(lat(0, 1), lat(1, 0))] = None
    _test_disconnected(lead)

# Test internal parts of magnetic_gauge

@pytest.mark.parametrize("shape",
                         [rectangle(5, 5), circle(4),
                          half_ring(5, 10)],
                         ids=['rectangle', 'circle', 'half-ring']
    )
@pytest.mark.parametrize("lattice", [square_lattice, honeycomb_lattice],
                         ids=['square', 'honeycomb'])
@pytest.mark.parametrize("neighbors", [1, 2, 3])
def test_minimal_cycle_basis(lattice, neighbors, shape):
    """Check that for lattice models on genus 0 shapes, nearly
       all loops have the same (minimal) length. This is not an
       equality, as there may be weird loops on the edges.
    """
    syst = Builder()
    syst.fill(model(lattice, neighbors), *shape)
    syst = syst.finalized()

    loops = gauge._loops_in_finite(syst)
    loop_counts = Counter(map(len, loops))
    min_loop = min(loop_counts)
    # arbitrarily allow 1% of slightly longer loops;
    # we don't make stronger guarantees about the quality
    # of our loop basis
    assert loop_counts[min_loop] / len(loops) > 0.99, loop_counts


def random_loop(n, max_radius=10, planar=False):
    """Return a loop of 'n' points.

    The loop is in the x-y plane if 'planar is False', otherwise
    each point is given a random perturbation in the z direction
    """
    theta = np.sort(2 * np.pi * np.random.rand(n))
    r = max_radius * np.random.rand(n)
    if planar:
        z = np.zeros((n,))
    else:
        z = 2 * (max_radius / 5) * (np.random.rand(n) - 1)
    return np.array([r * np.cos(theta), r * np.sin(theta), z]).transpose()


def test_constant_surface_integral():
    field_direction = np.random.rand(3)
    field_direction /= np.linalg.norm(field_direction)
    loop = random_loop(7)

    integral = gauge._surface_integral

    I = integral(lambda r: field_direction, loop)
    assert np.isclose(I, integral(field_direction, loop))
    assert np.isclose(I, integral(lambda r: field_direction, loop, average=True))


def circular_field(r_vec):
    return np.array([r_vec[1], -r_vec[0], 0])


def test_invariant_surface_integral():
    """Surface integral should be identical if we apply a random
       rotation to loop and vector field.
    """
    integral = gauge._surface_integral
    # loop with random orientation
    orig_loop = loop = random_loop(7)
    I = integral(circular_field, loop)
    for _ in range(4):
        rot = special_ortho_group.rvs(3)
        loop = orig_loop @ rot.transpose()
        assert np.isclose(I, integral(lambda r: rot @ circular_field(rot.transpose() @ r), loop))


@pytest.fixture
def system_and_gauge():
    def hopping(a, b, peierls):
        return -1 * peierls(a, b)

    syst = Builder()
    syst[(square_lattice(i, j) for i in range(3) for j in range(10))] = 4
    syst[square_lattice.neighbors()] = hopping

    lead = Builder(lattice.TranslationalSymmetry((-1, 0)))
    lead[(square_lattice(0, j) for j in range(10))] = 4
    lead[square_lattice.neighbors()] = hopping

    syst.attach_lead(lead.substituted(peierls='peierls_left'))
    syst.attach_lead(lead.reversed().substituted(peierls='peierls_right'))

    syst = syst.finalized()

    magnetic_gauge = gauge.magnetic_gauge(syst)

    return syst, magnetic_gauge


@pytest.mark.parametrize('B',[0, 0.1, lambda r: 0.1 * np.exp(-r[1]**2)])
def test_uniform_magnetic_field(system_and_gauge, B):
    syst, gauge = system_and_gauge

    peierls, peierls_left, peierls_right = gauge(B, B, B)

    params = dict(peierls=peierls, peierls_left=peierls_left,
                  peierls_right=peierls_right)

    s = kwant.smatrix(syst, energy=0.6, params=params)
    t = s.submatrix(1, 0)

    assert t.shape > (0, 0)  # sanity check
    assert np.allclose(np.abs(t)**2, np.eye(*t.shape))


def test_phase_sign(system_and_gauge):
    syst, gauge = system_and_gauge

    peierls, peierls_left, peierls_right = gauge(0.1, 0.1, 0.1)

    params = dict(peierls=peierls, peierls_left=peierls_left,
                  peierls_right=peierls_right)

    cut = [(square_lattice(1, j), square_lattice(0, j))
            for j in range(10)]
    J = kwant.operator.Current(syst, where=cut)
    J = J.bind(params=params)

    psi = kwant.wave_function(syst, energy=0.6, params=params)(0)[0]

    # Electrons incident from the left travel along the *top*
    # edge of the Hall bar in the presence of a magnetic field
    # out of the plane
    j = J(psi)
    j_bottom = sum(j[0:5])
    j_top = sum(j[5:10])

    assert np.isclose(j_top + j_bottom, 1)  # sanity check
    assert j_top > 0.9
